import pathlib

import numpy as np

from ggf.stress.boyde2009 import core


def test_basic():
    # Originally, the reference data ere created with Boyde's Matlab
    # script using default parameters. Due to a mistake in the boundary
    # function (issue #1), the reference data were recreated with the
    # Python ggf package.
    rpath = pathlib.Path(__file__).resolve().parent / "data"
    th_ref, sigmarr_ref = np.loadtxt(str(rpath / "stress_default.dat"))
    coeff_ref = np.loadtxt(str(rpath / "coeff_default.dat"))
    th, sigmarr, coeff = core.stress(ret_legendre_decomp=True)
    assert np.allclose(th, th_ref)
    assert np.allclose(sigmarr, sigmarr_ref)
    assert np.allclose(coeff, coeff_ref)


def test_basic_barton():
    # Originally, the reference data ere created with Boyde's Matlab
    # script using default parameters. Due to a mistake in the boundary
    # function (issue #1), the reference data were recreated with the
    # Python ggf package.
    rpath = pathlib.Path(__file__).resolve().parent / "data"
    th_ref, sigmarr_ref = np.loadtxt(str(rpath / "stress_default_barton.dat"))
    coeff_ref = np.loadtxt(str(rpath / "coeff_default_barton.dat"))
    th, sigmarr, coeff = core.stress(field_approx="barton",
                                     ret_legendre_decomp=True)
    assert np.allclose(th, th_ref)
    assert np.allclose(sigmarr, sigmarr_ref)
    assert np.allclose(coeff, coeff_ref)


def test_droplet():
    rpath = pathlib.Path(__file__).resolve().parent / "data"
    coeff_ref = np.loadtxt(str(rpath / "coeff_droplet1.dat"))
    semi_major = 4.62469e-6
    semi_minor = 4.589914e-6
    # We take the average of the two to compute the stress.
    # This is wrong.
    radius = (semi_major + semi_minor) / 2
    th, sigmarr, coeff = core.stress(semi_minor=radius,
                                     object_index=1.41,
                                     medium_index=1.3465,
                                     ret_legendre_decomp=True)
    assert np.allclose(coeff, coeff_ref)


def test_barton_davis_difference():
    th1, sigmarr1 = core.stress(semi_minor=1e-6, field_approx="davis")
    th2, sigmarr2 = core.stress(semi_minor=1e-6, field_approx="barton")
    assert np.allclose(sigmarr1, sigmarr2, rtol=0, atol=3e-4)


if __name__ == "__main__":
    test_droplet()
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import numpy as np

import ggf
from ggf.stress import boyde2009


def test_basic():
    """Test that all parameters are passed correctly"""
    radius = 2.8466e-6
    stretch_ratio = 0.1
    poisson_ratio = 0

    semi_minor = radius
    semi_major = radius * (1 + stretch_ratio)

    object_index = 1.3776
    medium_index = 1.335
    effective_fiber_distance = 100e-6
    mode_field_diameter = 4.8e-6
    wavelength = 800e-9
    power_per_fiber = .6

    n_points = 10

    th1, sg1 = boyde2009.core.stress(
        semi_minor=radius,
        stretch_ratio=stretch_ratio,
        poisson_ratio=poisson_ratio,
        object_index=object_index,
        medium_index=medium_index,
        dist=effective_fiber_distance/2,
        wavelength=wavelength,
        beam_waist=mode_field_diameter/2/wavelength,
        power_left=power_per_fiber,
        power_right=power_per_fiber,
        n_points=n_points)

    th2, sg2 = ggf.stress.get_stress(
        model="boyde2009",
        semi_major=semi_major,
        semi_minor=semi_minor,
        object_index=object_index,
        medium_index=medium_index,
        effective_fiber_distance=effective_fiber_distance,
        mode_field_diameter=mode_field_diameter,
        power_per_fiber=power_per_fiber,
        wavelength=wavelength,
        n_points=n_points)

    assert np.allclose(th1, th2)
    assert np.allclose(sg1, sg2)


def test_boundary():
    """Geometry should not depend on Poisson's ratio"""
    radius = 2.8466e-6
    stretch_ratio = 0.1
    poisson_ratio = 0.45

    # general formulas
    semi_minor = radius * (1 - poisson_ratio * stretch_ratio)
    semi_major = radius * (1 + stretch_ratio)

    costheta = np.cos(np.linspace(0, np.pi, 100))
    b1 = boyde2009.core.boundary(
        costheta=costheta,
        a=1,
        epsilon=(semi_major - semi_minor) / semi_minor,
        nu=0)
    b2 = boyde2009.core.boundary(
        costheta=costheta,
        a=1,
        epsilon=stretch_ratio,
        nu=poisson_ratio)

    assert np.allclose(b1, b2)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import numpy as np

from ggf import matlab_funcs


def test_besselj():
    # https://de.mathworks.com/help/matlab/ref/besselj.html
    z = np.arange(6) * .2
    req = matlab_funcs.besselj(1, z)
    ans = [0,
           0.0995,
           0.1960,
           0.2867,
           0.3688,
           0.4401,
           ]
    assert np.allclose(req, ans, rtol=0, atol=5e-5)


def test_besselh():
    z = np.arange(1, 6) * .2
    req = matlab_funcs.besselh(2, z)
    ansr = [0.0050,
            0.0197,
            0.0437,
            0.0758,
            0.1149,
            ]
    ansi = [-32.1571,
            -8.2983,
            -3.8928,
            -2.3586,
            -1.6507,
            ]
    assert np.allclose(req.real, ansr, rtol=0, atol=5e-5)
    assert np.allclose(req.imag, ansi, rtol=0, atol=5e-5)


def test_gammaln():
    x = np.array([1/5, 1/2, 2/3, 8/7, 3])
    req = matlab_funcs.gammaln(x)
    ans = [1.5241, 0.5724, 0.3032, -0.0667, 0.6931]
    assert np.allclose(req, ans, rtol=0, atol=5e-5)


def test_legendre():
    z = np.arange(6) * .2
    req = matlab_funcs.legendre(3, z)
    ans = [[0.000, -.2800, -.440, -.3600, .0800, 1.0000],
           [1.5000, 1.1758, 0.2750, -0.9600, -1.980, 0.0000],
           [0.0000, 2.8800, 5.0400, 5.7600, 4.3200, 0.0000],
           [-15.0000, -14.1091, -11.5481, -7.6800, -3.2400, 0.0000]
           ]
    assert np.allclose(req, ans, rtol=0, atol=5e-5)


def test_lscov():
    # https://de.mathworks.com/help/matlab/ref/lscov.html
    x1 = np.array([.2, .5, .6, .8, 1.0, 1.1])
    x2 = np.array([.1, .3, .4, .9, 1.1, 1.4])
    X = np.array([np.ones(x1.size), x1, x2]).T
    y = np.array([.17, .26, .28, .23, .27, .34])

    w = np.ones_like(x1)
    req = matlab_funcs.lscov(X, y, w)
    ans = [0.1203, 0.3284, -0.1312]
    assert np.allclose(req, ans, rtol=0, atol=5e-5)

    w2 = np.array([1, 1, 1, 1, 1, .1])
    req2 = matlab_funcs.lscov(X, y, w2)
    ans2 = [0.1046, 0.4614, -0.2621]
    assert np.allclose(req2, ans2, rtol=0, atol=5e-5)


def test_quadl():
    def myfunc(x): return 1./(x**3-2*x-5)
    Q = matlab_funcs.quadl(myfunc, 0, 2)
    assert np.allclose(Q, -0.4605, rtol=0, atol=5e-5)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



"""Other scientific functions"""
import numpy as np

from .matlab_funcs import legendre


def legendrePlm(m, l, x):
    a = np.concatenate((np.atleast_1d(np.zeros(m)),
                        np.atleast_1d(np.array(1)),
                        np.atleast_1d(np.zeros(l - m))))
    b = legendre(l, x)
    return np.real_if_close(np.dot(a, b)[0])



# flake8: noqa
from pkg_resources import resource_filename
from functools import lru_cache
import warnings

import numpy as np

from ...matlab_funcs import besselh, besselj, gammaln, lscov, quadl
from ...sci_funcs import legendrePlm
from ...core import stress2legendre


def boundary(costheta, a=1, epsilon=.1, nu=0):
    """Projected boundary of a prolate spheroid

    Compute the boundary according to equation (4) in
    :cite:`Boyde2009` with the addition of the
    Poisson's ratio of the object.

    .. math::

       B(\\theta) = a (1+\\epsilon)
       \\left[ (1+\\epsilon)^2 - \\epsilon (1+\\nu)
       (2+\\epsilon (1-\\nu)) \\cos^2 \\theta \\right]^{-1/2}

    This boundary function was derived for a prolate spheroid under
    the assumption that the semi-major axis :math:`a` and the
    semi-minor axes :math:`b=c` are defined as

    .. math::

       a = b \\cdot \\frac{1+ \\epsilon}{1- \\nu \\epsilon}

    The boundary function :math:`B(\\theta)` can be derived with
    the above relation using the equation for a prolate spheroid.

    Parameters
    ----------
    costheta: float or np.ndarray
        Cosine of polar coordinates :math:`\\theta`
        at which to compute the boundary.
    a: float
        Equatorial radii of prolate spheroid (semi-minor axis).
    epsilon: float
        Stretch ratio; defines size of semi-major axis:
        :math:`a = (1+\\epsilon) b`. Note that this is not
        the eccentricity of the prolate spheroid.
    nu: float
        Poisson's ratio :math:`\\nu` of the material.

    Returns
    -------
    B: 1d ndarray
        Radial object boundary in dependence of theta
        :math:`B(\\theta)`.

    Notes
    -----
    For :math:`\\nu=0`, the above equation becomes
    equation (4) in :cite:`Boyde2009`.
    """
    x = costheta
    B = a * (1 + epsilon) \
        / ((1 + epsilon)**2
           - epsilon * (1 + nu) * (2 + epsilon * (1 - nu)) * x**2)**.5
    return B


@lru_cache(maxsize=32)
def get_hgc():
    """Load hypergeometric coefficients from *hypergeomdata2.dat*.

    These coefficients were computed by Lars Boyde
    using Wolfram Mathematica.
    """
    hpath = resource_filename("ggf.stress.boyde2009", "hypergeomdata2.dat")
    hgc = np.loadtxt(hpath)
    return hgc


def stress(object_index=1.41, medium_index=1.3465, poisson_ratio=0.45,
           semi_minor=2.8466e-6, stretch_ratio=0.1, wavelength=780e-9,
           beam_waist=3, power_left=.6, power_right=.6, dist=100e-6,
           n_points=100, theta_max=np.pi, field_approx="davis",
           ret_legendre_decomp=False, verbose=False):
    """Compute the stress acting on a prolate spheroid

    The prolate spheroid has semi-major axis :math:`a` and
    semi-minor axis :math:`b=c`.

    Parameters
    ----------
    object_index: float
        Refractive index of the spheroid
    medium_index: float
        Refractive index of the surrounding medium
    poisson_ratio: float
        Poisson's ratio of the spheroid material
    semi_minor: float
        Semi-minor axis (inner) radius of the stretched object
        :math:`b=c`.
    stretch_ratio: float
        Measure of the deformation, defined as :math:`(a - b) / b`
    wavelength: float
        Wavelenth of the gaussian beam [m]
    beam_waist: float
        Beam waist radius of the gaussian beam [wavelengths]
    power_left: float
        Laser power of the left beam [W]
    power_right: float
        Laser power of the right beam [W]
    dist: float
        Distance between beam waist and object center [m]
    n_points: int
        Number of points to compute stresses for
    theta_max: float
        Maximum angle to compute stressed for
    field_approx: str
        TODO
    ret_legendre_decomp: bool
        If True, return coefficients of decomposition of stress
        into Legendre polynomials
    verbose: int
        Increase verbosity

    Returns
    -------
    theta: 1d ndarray
        Angles for which stresses are computed
    sigma_rr: 1d ndarray
        Radial stress corresponding to angles
    coeff: 1d ndarray
        If `ret_legendre_decomp` is True, return compositions
        of decomposition of stress into Legendre polynomials.

    Notes
    -----
    - The angles `theta` are computed on a grid that does not
      include zero and `theta_max`.
    - This implementation was first presented in :cite:`Boyde2009`.
    """
    if field_approx not in ["davis", "barton"]:
        raise ValueError("`field_approx` must be 'davis' or 'barton'")
    object_index = complex(object_index)
    medium_index = complex(medium_index)
    W0 = beam_waist * wavelength
    epsilon = stretch_ratio
    nu = poisson_ratio

    # ZRL = 0.5*medium_index*2*np.pi/wavelength*W0**2 # Rayleigh range [m]
    # WZ  = W0*(1+(beam_pos+d)**2/ZRL**2)**0.5   # beam waist at specified
    # position [m]

    K0 = 2 * np.pi / wavelength         # wave vector [m]
    Alpha = semi_minor * K0                # size parameter
    C = 3e8                 # speed of light [m/s]

    # maximum number of orders
    lmax = np.int(np.round(2 + Alpha + 4 * (Alpha)**(1 / 3) + 10))
    if lmax > 120:
        msg = 'Required number of orders for accurate expansion exceeds allowed maximum!' \
              + 'Reduce size of trapped particle!'
        raise ValueError(msg)

    if epsilon == 0:
        # spherical object, no point-matching needed (mmax = 0)
        mmax = 3
    else:
        if (epsilon > 0.15):
            warnings.warn(
                'Warning, cell stretching ratio is high: {}'.format(epsilon))
        # spheroidal object, point-matching required (mmax has to be divisible
        # by 3)
        mmax = 6 * lmax

    # permittivity in surrounding medium [1]
    EpsilonI = medium_index**2
    EpsilonII = object_index**2          # permittivity in within cell [1]

    MuI = 1.000               # permeability in surrounding medium [1]
    MuII = 1.000               # permeability within cell [1]
    # wave constant in Maxwell's equations (surrounding medium) [1/m]
    K1I = 1j * K0 * EpsilonI
    # wave constant in Maxwell's equations (within cell) [1/m]
    K1II = 1j * K0 * EpsilonII
    # wave constant in Maxwell's equations (surrounding medium) [1/m]
    K2I = 1j * K0
    # wave constant in Maxwell's equations (within cell) [1/m]
    K2II = 1j * K0
    KI = (-K1I * K2I)**0.5       # wave vector (surrounding medium) [1/m]
    KII = (-K1II * K2II)**0.5    # wave vector (within cell) [1/m]

    # dimensionless parameters
    k0 = 1                    # wave vector
    a = semi_minor * K0               # internal radius of stretched cell
    d = dist * K0               # distance from cell centre to optical stretcher
    # ap = a*(1+stretch_ratio)             # semi-major axis (after stretching)
    # bp = a*(1-poisson_ratio*stretch_ratio)          # semi-minor axis (after
    # stretching)
    w0 = W0 * K0              # Gaussian width
    # wave constant in Maxwell's equations (surrounding medium)
    k1I = K1I / K0
    # wave constant in Maxwell's equations (within cell)
    k1II = K1II / K0
    # wave constant in Maxwell's equations (surrounding medium)
    k2I = K2I / K0
    # wave constant in Maxwell's equations (within cell)
    k2II = K2II / K0
    kI = KI / K0               # wave vector (surrounding medium)
    kII = KII / K0              # wave vector (within cell)
    beta = kI                  # wave vector of Gaussian beam

    # other definitions
    # amplitude of electric field of left laser [kg m/(s**2 C)]
    EL = np.sqrt(power_left / (medium_index * C * W0**2))
    # amplitude of electric field of right laser [kg m/(s**2 C)]
    ER = np.sqrt(power_right / (medium_index * C * W0**2))
    HL = beta / k0 * EL            # left laser amplitude of magnetic field
    HR = beta / k0 * ER            # right laser amplitude of magnetic field

    zR = beta * w0**2 / 2          # definition of Rayleigh range
    S = (1 + 1j * d / zR)**(-1)     # complex amplitude for Taylor expansion
    s = 1 / (beta * w0)           # expansion parameter for Gaussian (Barton)

    # Functions
    # object boundary function: r(th) = a*B1(x) x= cos(th)
    def B1(x): return boundary(costheta=x, a=1, epsilon=epsilon, nu=nu)

    # Riccati Bessel functions and their derivatives
    # Riccati Bessel function (psi)
    def psi(l, z): return (np.pi / 2 * z)**(1 / 2) * besselj(l + 1 / 2, z)

    def psi1(l, z): return (np.pi / (2. * z))**(1 / 2) * \
        (z * besselj(l - 1 / 2, z) - l *
         besselj(l + 1 / 2, z))  # first derivative (psi')

    def psi2(l, z): return (np.pi / 2)**(1 / 2) * (l + l**2 - z**2) * \
        besselj(l + 1 / 2, z) * z**(-3 / 2)        # second derivative (psi'')

    # First order Taylor expansion of psi is too inaccurate for larger values of k*a*Eps.
    # Hence, to match 1-st and higher order terms in Eps, subtract the 0-th order terms (no angular dependence)
    # from the exact function psi (including angular dependence)
    # Riccati Bessel function excluding angular dependence in 0-th order
    # (psiex)
    def psiex(l, z, x): return psi(l, z * B1(x)) - psi(l, z)

    def psi1ex(l, z, x): return psi1(l, z * B1(x)) - \
        psi1(l, z)  # first derivative of psiex

    def psi2ex(l, z, x): return psi2(l, z * B1(x)) - \
        psi2(l, z)  # second derivative of psi

    # defined for abbreviation
    def psixx(l, z, x): return psi(l, z * B1(x))

    def psi1xx(l, z, x): return psi1(l, z * B1(x))

    def psi2xx(l, z, x): return psi2(l, z * B1(x))

    # Hankel function and its derivative
    def xi(l, z): return (np.pi / 2 * z)**(1 / 2) * besselh(l + 1 / 2, z)

    def xi1(l, z): return (np.pi / (2 * z))**(1 / 2) * \
        ((l + 1) * besselh(l + 1 / 2, z) - z * besselh(l + 3 / 2, z))

    def xi2(l, z): return (np.pi / 2)**(1 / 2) / \
        z**(3 / 2) * (l + l**2 - z**2) * besselh(l + 1 / 2, z)

    # Comments: see above for psiex
    def xiex(l, z, x): return xi(l, z * B1(x)) - xi(l, z)

    def xi1ex(l, z, x): return xi1(l, z * B1(x)) - xi1(l, z)

    def xi2ex(l, z, x): return xi2(l, z * B1(x)) - xi2(l, z)

    def xixx(l, z, x): return xi(l, z * B1(x))

    def xi1xx(l, z, x): return xi1(l, z * B1(x))

    def xi2xx(l, z, x): return xi2(l, z * B1(x))

    #% Associated Legendre functions P(m)_l(x) and their derivatives
    #% select mth component of vector 'legendre'  [P**(m)_l(x)]
    # [zeros(m,1);1;zeros(l-m,1)].'*legendre(l,x)

    #% legendre polynomial [P**(m)_l(x)]
    def legendrePl(l, x): return legendrePlm(1, l, x)
    #% legendre polynomial [P**(1)_l(x)]

    def legendrePlm1(m, l, x): return (
        (l - m + 1.) * legendrePlm(m, l + 1, x) - (l + 1.) * x * legendrePlm(m, l, x)) / (x**2 - 1)
    #% derivative d/dx[P**(m)_l(x)]

    def legendrePl1(l, x): return legendrePlm1(1, l, x)

    # defined to avoid division by zero (which can occur for x=1 in
    # legendrePl1...
    def legendrePlmex1(m, l, x): return -((l - m + 1) *
                                          legendrePlm(m, l + 1, x) - (l + 1) * x * legendrePlm(m, l, x))

    def legendrePlex1(l, x): return legendrePlmex1(1, l, x)

    # Hypergeometric and Gamma functions
    hypergeomcoeff = get_hgc()

    ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ## ##
    # Gaussian beam (incident fields) - in Cartesian basis (either Davis first order or Barton fifth order fields)
    # electric and magnetic fields according to Davis (first order)
    if field_approx == "davis":
        # left
        def eExiL(r, th, phi): return EL * (1 + 1j * (r * np.cos(th) + d) / zR)**(-1) * np.exp(-r**2 *
                                                                                               np.sin(th)**2 / (w0**2 * (1 + 1j * (r * np.cos(th) + d) / zR))) * np.exp(1j * beta * (r * np.cos(th) + d))

        def eEyiL(r, th, phi): return 0

        def eEziL(r, th, phi): return -1j * (1 + 1j * (r * np.cos(th) + d) /
                                             zR)**(-1) * r * np.sin(th) * np.cos(phi) / zR * eExiL(r, th, phi)

        def eHxiL(r, th, phi): return 0

        def eHyiL(r, th, phi): return HL * (1 + 1j * (r * np.cos(th) + d) / zR)**(-1) * np.exp(-r**2 *
                                                                                               np.sin(th)**2 / (w0**2 * (1 + 1j * (r * np.cos(th) + d) / zR))) * np.exp(1j * beta * (r * np.cos(th) + d))

        def eHziL(r, th, phi): return -1j * (1 + 1j * (r * np.cos(th) + d) /
                                             zR)**(-1) * r * np.sin(th) * np.sin(phi) / zR * eHyiL(r, th, phi)

        # right
        def eExiR(r, th, phi): return ER * (1 - 1j * (r * np.cos(th) - d) / zR)**(-1) * np.exp(-r**2 *
                                                                                               np.sin(th)**2 / (w0**2 * (1 - 1j * (r * np.cos(th) - d) / zR))) * np.exp(-1j * beta * (r * np.cos(th) - d))

        def eEyiR(r, th, phi): return 0

        def eEziR(r, th, phi): return +1j * (1 - 1j * (r * np.cos(th) - d) /
                                             zR)**(-1) * r * np.sin(th) * np.cos(phi) / zR * eExiR(r, th, phi)

        def eHxiR(r, th, phi): return 0

        def eHyiR(r, th, phi): return -HR * (1 - 1j * (r * np.cos(th) - d) / zR)**(-1) * np.exp(-r**2 *
                                                                                                np.sin(th)**2 / (w0**2 * (1 - 1j * (r * np.cos(th) - d) / zR))) * np.exp(-1j * beta * (r * np.cos(th) - d))

        def eHziR(r, th, phi): return +1j * (1 - 1j * (r * np.cos(th) - d) /
                                             zR)**(-1) * r * np.sin(th) * np.sin(phi) / zR * eHyiR(r, th, phi)

    else:  # electric and magnetic fields according to Barton (fifth order)
        # Note: in Barton propagation is: np.exp(i (omega*t-k*z)) and not np.exp(i (k*z-omega*t)).
        # Hence, take complex conjugate of equations or make the changes z ->
        # -z, Ez -> -Ez, Hz -> -Hz while x and y are not affected.
        def Xi(r, th, phi): return r * np.sin(th) * np.cos(phi) / w0

        def Eta(r, th, phi): return r * np.sin(th) * np.sin(phi) / w0

        def ZetaL(r, th): return (r * np.cos(th) + d) / (beta * w0**2)

        def Rho(r, th, phi): return np.sqrt(
            Xi(r, th, phi)**2 + Eta(r, th, phi)**2)

        def QL(r, th): return 1. / (1j + 2 * ZetaL(r, th))

        def Psi0L(r, th, phi): return 1j * QL(r, th) * \
            np.exp(-1j * (Rho(r, th, phi))**2 * QL(r, th))

        def eExiL(r, th, phi): return np.conj(EL * Psi0L(r, th, phi) * np.exp(-1j * ZetaL(r, th) / s**2) *
                                              (1 + s**2 * (-Rho(r, th, phi)**2 * QL(r, th)**2 + 1j * Rho(r, th, phi)**4 * QL(r, th)**3 - 2 * QL(r, th)**2 * Xi(r, th, phi)**2) +
                                               s**4 * (2 * Rho(r, th, phi)**4 * QL(r, th)**4 - 3 * 1j * Rho(r, th, phi)**6 * QL(r, th)**5 - 0.5 * Rho(r, th, phi)**8 * QL(r, th)**6 +
                                                       (8 * Rho(r, th, phi)**2 * QL(r, th)**4 - 2 * 1j * Rho(r, th, phi)**4 * QL(r, th)**5) * Xi(r, th, phi)**2)))

        def eEyiL(r, th, phi): return np.conj(EL * Psi0L(r, th, phi) * np.exp(-1j * ZetaL(r, th) / s**2) *
                                              (s**2 * (-2 * QL(r, th)**2 * Xi(r, th, phi) * Eta(r, th, phi)) +
                                               s**4 * ((8 * Rho(r, th, phi)**2 * QL(r, th)**4 - 2 * 1j * Rho(r, th, phi)**4 * QL(r, th)**5) * Xi(r, th, phi) * Eta(r, th, phi))))

        def eEziL(r, th, phi): return np.conj(EL * Psi0L(r, th, phi) * np.exp(-1j * ZetaL(r, th) / s**2) *
                                              (s * (-2 * QL(r, th) * Xi(r, th, phi)) + s**3 * (6 * Rho(r, th, phi)**2 * QL(r, th)**3 - 2 * 1j * Rho(r, th, phi)**4 * QL(r, th)**4) * Xi(r, th, phi) +
                                               s**5 * (-20 * Rho(r, th, phi)**4 * QL(r, th)**5 + 10 * 1j * Rho(r, th, phi)**6 * QL(r, th)**6 + Rho(r, th, phi)**8 * QL(r, th)**7) * Xi(r, th, phi)))

        def eHxiL(r, th, phi): return np.conj(+np.sqrt(EpsilonI) * EL * Psi0L(r, th, phi) * np.exp(-1j * ZetaL(r, th) / s**2) *
                                              (s**2 * (-2 * QL(r, th)**2 * Xi(r, th, phi) * Eta(r, th, phi)) +
                                               s**4 * ((8 * Rho(r, th, phi)**2 * QL(r, th)**4 - 2 * 1j * Rho(r, th, phi)**4 * QL(r, th)**5) * Xi(r, th, phi) * Eta(r, th, phi))))

        def eHyiL(r, th, phi): return np.conj(+np.sqrt(EpsilonI) * EL * Psi0L(r, th, phi) * np.exp(-1j * ZetaL(r, th) / s**2) *
                                              (1 + s**2 * (-Rho(r, th, phi)**2 * QL(r, th)**2 + 1j * Rho(r, th, phi)**4 * QL(r, th)**3 - 2 * QL(r, th)**2 * Eta(r, th, phi)**2) +
                                               s**4 * (2 * Rho(r, th, phi)**4 * QL(r, th)**4 - 3 * 1j * Rho(r, th, phi)**6 * QL(r, th)**5 - 0.5 * Rho(r, th, phi)**8 * QL(r, th)**6 +
                                                       (8 * Rho(r, th, phi)**2 * QL(r, th)**4 - 2 * 1j * Rho(r, th, phi)**4 * QL(r, th)**5) * Eta(r, th, phi)**2)))

        def eHziL(r, th, phi): return np.conj(np.sqrt(EpsilonI) * EL * Psi0L(r, th, phi) * np.exp(-1j * ZetaL(r, th) / s**2) *
                                              (s * (-2 * QL(r, th) * Eta(r, th, phi)) + s**3 * (6 * Rho(r, th, phi)**2 * QL(r, th)**3 - 2 * 1j * Rho(r, th, phi)**4 * QL(r, th)**4) * Eta(r, th, phi) +
                                               s**5 * (-20 * Rho(r, th, phi)**4 * QL(r, th)**5 + 10 * 1j * Rho(r, th, phi)**6 * QL(r, th)**6 + Rho(r, th, phi)**8 * QL(r, th)**7) * Eta(r, th, phi)))

        # right
        # Take left fiber fields and make coordinate changes (x,y,z,d) ->
        # (x,-y,-z,d) and amplitude changes (Ex,Ey,Ez) -> (Ex,-Ey,-Ez).
        def ZetaR(r, th): return -(r * np.cos(th) - d) / (beta * w0**2)

        def QR(r, th): return 1. / (1j + 2 * ZetaR(r, th))

        def Psi0R(r, th, phi): return 1j * QR(r, th) * \
            np.exp(-1j * (Rho(r, th, phi))**2 * QR(r, th))

        def eExiR(r, th, phi): return np.conj(ER * Psi0R(r, th, phi) * np.exp(-1j * ZetaR(r, th) / s**2) *
                                              (1 + s**2 * (-Rho(r, th, phi)**2 * QR(r, th)**2 + 1j * Rho(r, th, phi)**4 * QR(r, th)**3 - 2 * QR(r, th)**2 * Xi(r, th, phi)**2) +
                                               s**4 * (2 * Rho(r, th, phi)**4 * QR(r, th)**4 - 3 * 1j * Rho(r, th, phi)**6 * QR(r, th)**5 - 0.5 * Rho(r, th, phi)**8 * QR(r, th)**6 +
                                                       (8 * Rho(r, th, phi)**2 * QR(r, th)**4 - 2 * 1j * Rho(r, th, phi)**4 * QR(r, th)**5) * Xi(r, th, phi)**2)))

        def eEyiR(r, th, phi): return np.conj(ER * Psi0R(r, th, phi) * np.exp(-1j * ZetaR(r, th) / s**2) *
                                              (s**2 * (-2 * QR(r, th)**2 * Xi(r, th, phi) * Eta(r, th, phi)) +
                                               s**4 * ((8 * Rho(r, th, phi)**2 * QR(r, th)**4 - 2 * 1j * Rho(r, th, phi)**4 * QR(r, th)**5) * Xi(r, th, phi) * Eta(r, th, phi))))

        def eEziR(r, th, phi): return - np.conj(ER * Psi0R(r, th, phi) * np.exp(-1j * ZetaR(r, th) / s**2) *
                                                (s * (-2 * QR(r, th) * Xi(r, th, phi)) + s**3 * (6 * Rho(r, th, phi)**2 * QR(r, th)**3 - 2 * 1j * Rho(r, th, phi)**4 * QR(r, th)**4) * Xi(r, th, phi) +
                                                 s**5 * (-20 * Rho(r, th, phi)**4 * QR(r, th)**5 + 10 * 1j * Rho(r, th, phi)**6 * QR(r, th)**6 + Rho(r, th, phi)**8 * QR(r, th)**7) * Xi(r, th, phi)))

        def eHxiR(r, th, phi): return - np.conj(+np.sqrt(EpsilonI) * ER * Psi0R(r, th, phi) * np.exp(-1j * ZetaR(r, th) / s**2) *
                                                (s**2 * (-2 * QR(r, th)**2 * Xi(r, th, phi) * Eta(r, th, phi)) +
                                                 s**4 * ((8 * Rho(r, th, phi)**2 * QR(r, th)**4 - 2 * 1j * Rho(r, th, phi)**4 * QR(r, th)**5) * Xi(r, th, phi) * Eta(r, th, phi))))

        def eHyiR(r, th, phi): return - np.conj(+np.sqrt(EpsilonI) * ER * Psi0R(r, th, phi) * np.exp(-1j * ZetaR(r, th) / s**2) *
                                                (1 + s**2 * (-Rho(r, th, phi)**2 * QR(r, th)**2 + 1j * Rho(r, th, phi)**4 * QR(r, th)**3 - 2 * QR(r, th)**2 * Eta(r, th, phi)**2) +
                                                 s**4 * (2 * Rho(r, th, phi)**4 * QR(r, th)**4 - 3 * 1j * Rho(r, th, phi)**6 * QR(r, th)**5 - 0.5 * Rho(r, th, phi)**8 * QR(r, th)**6 +
                                                         (8 * Rho(r, th, phi)**2 * QR(r, th)**4 - 2 * 1j * Rho(r, th, phi)**4 * QR(r, th)**5) * Eta(r, th, phi)**2)))

        def eHziR(r, th, phi): return np.conj(np.sqrt(EpsilonI) * ER * Psi0R(r, th, phi) * np.exp(-1j * ZetaR(r, th) / s**2) *
                                              (s * (-2 * QR(r, th) * Eta(r, th, phi)) + s**3 * (6 * Rho(r, th, phi)**2 * QR(r, th)**3 - 2 * 1j * Rho(r, th, phi)**4 * QR(r, th)**4) * Eta(r, th, phi) +
                                               s**5 * (-20 * Rho(r, th, phi)**4 * QR(r, th)**5 + 10 * 1j * Rho(r, th, phi)**6 * QR(r, th)**6 + Rho(r, th, phi)**8 * QR(r, th)**7) * Eta(r, th, phi)))

    # Gaussian beam (incident fields) - in spherical polar coordinates basis
    # left
    def eEriL(r, th, phi): return np.sin(th) * np.cos(phi) * eExiL(r, th, phi) + \
        np.sin(th) * np.sin(phi) * eEyiL(r, th, phi) + \
        np.cos(th) * eEziL(r, th, phi)

    def eEthiL(r, th, phi): return np.cos(th) * np.cos(phi) * eExiL(r, th, phi) + \
        np.cos(th) * np.sin(phi) * eEyiL(r, th, phi) - \
        np.sin(th) * eEziL(r, th, phi)

    def eEphiiL(r, th, phi): return - np.sin(phi) * \
        eExiL(r, th, phi) + np.cos(phi) * eEyiL(r, th, phi)

    def eHriL(r, th, phi): return np.sin(th) * np.cos(phi) * eHxiL(r, th, phi) + \
        np.sin(th) * np.sin(phi) * eHyiL(r, th, phi) + \
        np.cos(th) * eHziL(r, th, phi)

    def eHthiL(r, th, phi): return np.cos(th) * np.cos(phi) * eHxiL(r, th, phi) + \
        np.cos(th) * np.sin(phi) * eHyiL(r, th, phi) - \
        np.sin(th) * eHziL(r, th, phi)

    def eHphiiL(r, th, phi): return - np.sin(phi) * \
        eHxiL(r, th, phi) + np.cos(phi) * eHyiL(r, th, phi)
    # right

    def eEriR(r, th, phi): return np.sin(th) * np.cos(phi) * eExiR(r, th, phi) + \
        np.sin(th) * np.sin(phi) * eEyiR(r, th, phi) + \
        np.cos(th) * eEziR(r, th, phi)

    def eEthiR(r, th, phi): return np.cos(th) * np.cos(phi) * eExiR(r, th, phi) + \
        np.cos(th) * np.sin(phi) * eEyiR(r, th, phi) - \
        np.sin(th) * eEziR(r, th, phi)

    def eEphiiR(r, th, phi): return - np.sin(phi) * \
        eExiR(r, th, phi) + np.cos(phi) * eEyiR(r, th, phi)

    def eHriR(r, th, phi): return np.sin(th) * np.cos(phi) * eHxiR(r, th, phi) + \
        np.sin(th) * np.sin(phi) * eHyiR(r, th, phi) + \
        np.cos(th) * eHziR(r, th, phi)

    def eHthiR(r, th, phi): return np.cos(th) * np.cos(phi) * eHxiR(r, th, phi) + \
        np.cos(th) * np.sin(phi) * eHyiR(r, th, phi) - \
        np.sin(th) * eHziR(r, th, phi)

    def eHphiiR(r, th, phi): return - np.sin(phi) * \
        eHxiR(r, th, phi) + np.cos(phi) * eHyiR(r, th, phi)

    eBL = np.zeros(lmax, dtype=complex)
    mBL = np.zeros(lmax, dtype=complex)
    eBR = np.zeros(lmax, dtype=complex)
#    eBR_test = np.zeros(lmax, dtype=complex)
    mBR = np.zeros(lmax, dtype=complex)

    if field_approx == "davis":  # Davis
        tau = np.zeros(lmax, dtype=complex)
        for ii in range(lmax):
            l = ii + 1
            tau[ii] = 0                             # sum over m and n
            niimax = int(np.floor((l - 1) / 3))
            for n in range(niimax + 1):
                miimax = int(np.floor((l - 1) / 2 - 3 / 2 * n))
                for m in range(miimax + 1):
                    if l < (2 * m + 3 * n + 2):
                        Delta = 0
                    else:
                        Delta = 1
                    tau[ii] = tau[ii] + np.exp(gammaln(l / 2 - m - n) + gammaln(m + n + 2) - gammaln(l - 2 * m - 3 * n) - gammaln(l / 2 + 2) - gammaln(m + 1) - gammaln(n + 1)) * hypergeomcoeff[l - 1, m + n] \
                        * (-1)**(m + 1) / (S**m * (beta * zR)**n) * (S / (1j * beta * w0))**(2 * m + 2 * n) * (1 - Delta * 2 * S / (beta * zR) * (l - 1 - 2 * m - 3 * n))
#                    print(tau)

        # calculate expansion coefficients of order l for electric and magnetic
        # Debye potentials
        def emB(l): return S * np.exp(1j * beta * d) * (1j * beta / (2 * kI))**(l - 1) * \
            (l + 1 / 2)**2 / (l + 1) * \
            np.exp(gammaln(2 * l) - gammaln(l + 1)) * tau[l - 1]
        for ii in range(lmax):
            l = ii + 1
            # left
            eBL[ii] = EL * emB(l)
            mBL[ii] = HL * emB(l)
            # right
            # should include factor of (-1)**(l-1) for symmetry reasons, this
            # is taken into account only after least square fitting (see below)
            eBR[ii] = ER * emB(l)
            # should include factor of (-1)**l for symmetry reasons, this is
            # taken into account only after least square fitting (see
            mBR[ii] = HR * emB(l)
    else:  # Barton

        for ii in range(lmax):
            l = ii + 1
            eBL[ii] = np.sqrt(2) * quadl(lambda th: eEriL(a, th, np.pi / 4) * legendrePl(l, np.cos(
                th)) * np.sin(th), 0, np.pi) * (2 * l + 1) / (2 * l * (l + 1)) * a**2 / psi(l, kI * a) * kI**2 / (l * (l + 1))
            mBL[ii] = np.sqrt(2) * quadl(lambda th: eHriL(a, th, np.pi / 4) * legendrePl(l, np.cos(
                th)) * np.sin(th), 0, np.pi) * (2 * l + 1) / (2 * l * (l + 1)) * a**2 / psi(l, kI * a) * kI**2 / (l * (l + 1))
            eBR[ii] = np.sqrt(2) * quadl(lambda th: eEriR(a, th, np.pi / 4) * legendrePl(l, np.cos(
                th)) * np.sin(th), 0, np.pi) * (2 * l + 1) / (2 * l * (l + 1)) * a**2 / psi(l, kI * a) * kI**2 / (l * (l + 1))
            mBR[ii] = np.sqrt(2) * quadl(lambda th: eHriR(a, th, np.pi / 4) * legendrePl(l, np.cos(
                th)) * np.sin(th), 0, np.pi) * (2 * l + 1) / (2 * l * (l + 1)) * a**2 / psi(l, kI * a) * kI**2 / (l * (l + 1))

            # make symetrical with left expansion coefficients
            eBR[ii] = eBR[ii] * (-1)**(l - 1)
            # make symetrical with left expansion coefficients
            mBR[ii] = mBR[ii] * (-1)**l

    # coefficients for internal fields (eCl, mCl) and scattered fields (eDl,
    # mDl)
    eCL = np.zeros(lmax, dtype=complex)
    mCL = np.zeros(lmax, dtype=complex)
    eCR = np.zeros(lmax, dtype=complex)
    mCR = np.zeros(lmax, dtype=complex)

    eDL = np.zeros(lmax, dtype=complex)
    mDL = np.zeros(lmax, dtype=complex)
    eDR = np.zeros(lmax, dtype=complex)
    mDR = np.zeros(lmax, dtype=complex)
    for ii in range(lmax):
        l = ii + 1
        # internal (left and right)
        eCL[ii] = k1I / kI * (kII)**2 * (xi(l, kI * a) * psi1(l, kI * a) - xi1(l, kI * a) * psi(l, kI * a)) / (
            k1I * kII * xi(l, kI * a) * psi1(l, kII * a) - k1II * kI * xi1(l, kI * a) * psi(l, kII * a)) * eBL[ii]
        mCL[ii] = k2I / kI * (kII)**2 * (xi(l, kI * a) * psi1(l, kI * a) - xi1(l, kI * a) * psi(l, kI * a)) / (
            k2I * kII * xi(l, kI * a) * psi1(l, kII * a) - k2II * kI * xi1(l, kI * a) * psi(l, kII * a)) * mBL[ii]
        eCR[ii] = k1I / kI * (kII)**2 * (xi(l, kI * a) * psi1(l, kI * a) - xi1(l, kI * a) * psi(l, kI * a)) / (
            k1I * kII * xi(l, kI * a) * psi1(l, kII * a) - k1II * kI * xi1(l, kI * a) * psi(l, kII * a)) * eBR[ii]
        mCR[ii] = k2I / kI * (kII)**2 * (xi(l, kI * a) * psi1(l, kI * a) - xi1(l, kI * a) * psi(l, kI * a)) / (
            k2I * kII * xi(l, kI * a) * psi1(l, kII * a) - k2II * kI * xi1(l, kI * a) * psi(l, kII * a)) * mBR[ii]
        # scattered (left and right)
        eDL[ii] = (k1I * kII * psi(l, kI * a) * psi1(l, kII * a) - k1II * kI * psi1(l, kI * a) * psi(l, kII * a)) / \
            (k1II * kI * xi1(l, kI * a) * psi(l, kII * a) - k1I *
             kII * xi(l, kI * a) * psi1(l, kII * a)) * eBL[ii]
        mDL[ii] = (k2I * kII * psi(l, kI * a) * psi1(l, kII * a) - k2II * kI * psi1(l, kI * a) * psi(l, kII * a)) / \
            (k2II * kI * xi1(l, kI * a) * psi(l, kII * a) - k2I *
             kII * xi(l, kI * a) * psi1(l, kII * a)) * mBL[ii]
        eDR[ii] = (k1I * kII * psi(l, kI * a) * psi1(l, kII * a) - k1II * kI * psi1(l, kI * a) * psi(l, kII * a)) / \
            (k1II * kI * xi1(l, kI * a) * psi(l, kII * a) - k1I *
             kII * xi(l, kI * a) * psi1(l, kII * a)) * eBR[ii]
        mDR[ii] = (k2I * kII * psi(l, kI * a) * psi1(l, kII * a) - k2II * kI * psi1(l, kI * a) * psi(l, kII * a)) / \
            (k2II * kI * xi1(l, kI * a) * psi(l, kII * a) - k2I *
             kII * xi(l, kI * a) * psi1(l, kII * a)) * mBR[ii]

    # First Order Expansion Coefficients
    # coefficients for internal fields (eCcl, mCcl) and scattered fields
    # (eDdl, mDdl)
    eLambda1L = {}
    mLambda1L = {}
    eLambda2L = {}
    mLambda2L = {}
    eLambda3L = {}
    mLambda3L = {}
    eLambda1R = {}
    mLambda1R = {}
    eLambda2R = {}
    mLambda2R = {}
    eLambda3R = {}
    mLambda3R = {}

    for jj in range(lmax):
        l = jj + 1
        # left
        eLambda1L[l] = lambda x, l=l, jj=jj: (eBL[jj] / kI * psi1ex(l, kI * a, x) - eCL[jj] / kII * psi1ex(
            l, kII * a, x) + eDL[jj] / kI * xi1ex(l, kI * a, x))                   # electric parameter1 left
        mLambda1L[l] = lambda x, l=l, jj=jj: (mBL[jj] / kI * psi1ex(l, kI * a, x) - mCL[jj] / kII * psi1ex(
            l, kII * a, x) + mDL[jj] / kI * xi1ex(l, kI * a, x))                   # magnetic parameter1 left
        eLambda2L[l] = lambda x, l=l, jj=jj: (k1I / kI**2 * eBL[jj] * psiex(l, kI * a, x) - k1II / kII**2 * eCL[jj] * psiex(
            l, kII * a, x) + k1I / kI**2 * eDL[jj] * xiex(l, kI * a, x))   # electric parameter2 left
        mLambda2L[l] = lambda x, l=l, jj=jj: (k2I / kI**2 * mBL[jj] * psiex(l, kI * a, x) - k2II / kII**2 * mCL[jj] * psiex(
            l, kII * a, x) + k2I / kI**2 * mDL[jj] * xiex(l, kI * a, x))   # magnetic parameter2 left
        eLambda3L[l] = lambda x, l=l, jj=jj: (eBL[jj] * (psiex(l, kI * a, x) + psi2ex(l, kI * a, x)) * k1I - eCL[jj] * (psiex(l, kII * a, x) + psi2ex(l, kII * a, x)) * k1II
                                              + eDL[jj] * (xiex(l, kI * a, x) + xi2ex(l, kI * a, x)) * k1I)                                                       # electric parameter3 left
        mLambda3L[l] = lambda x, l=l, jj=jj: (mBL[jj] * (psiex(l, kI * a, x) + psi2ex(l, kI * a, x)) * MuI - mCL[jj] * (psiex(l, kII * a, x) + psi2ex(l, kII * a, x)) * MuII
                                              + mDL[jj] * (xiex(l, kI * a, x) + xi2ex(l, kI * a, x)) * MuI)                                                       # magnetic parameter3 left
        # right
        eLambda1R[l] = lambda x, l=l, jj=jj: (eBR[jj] / kI * psi1ex(l, kI * a, x) - eCR[jj] / kII * psi1ex(
            l, kII * a, x) + eDR[jj] / kI * xi1ex(l, kI * a, x))                   # electric parameter1 right
        mLambda1R[l] = lambda x, l=l, jj=jj: (mBR[jj] / kI * psi1ex(l, kI * a, x) - mCR[jj] / kII * psi1ex(
            l, kII * a, x) + mDR[jj] / kI * xi1ex(l, kI * a, x))                   # magnetic parameter1 right
        eLambda2R[l] = lambda x, l=l, jj=jj: (k1I / kI**2 * eBR[jj] * psiex(l, kI * a, x) - k1II / kII**2 * eCR[jj] * psiex(
            l, kII * a, x) + k1I / kI**2 * eDR[jj] * xiex(l, kI * a, x))   # electric parameter2 right
        mLambda2R[l] = lambda x, l=l, jj=jj: (k2I / kI**2 * mBR[jj] * psiex(l, kI * a, x) - k2II / kII**2 * mCR[jj] * psiex(
            l, kII * a, x) + k2I / kI**2 * mDR[jj] * xiex(l, kI * a, x))   # magnetic parameter2 right
        eLambda3R[l] = lambda x, l=l, jj=jj: (eBR[jj] * (psiex(l, kI * a, x) + psi2ex(l, kI * a, x)) * k1I - eCR[jj] * (psiex(l, kII * a, x) + psi2ex(l, kII * a, x)) * k1II
                                              + eDR[jj] * (xiex(l, kI * a, x) + xi2ex(l, kI * a, x)) * k1I)                                                       # electric parameter3 right
        mLambda3R[l] = lambda x, l=l, jj=jj: (mBR[jj] * (psiex(l, kI * a, x) + psi2ex(l, kI * a, x)) * MuI - mCR[jj] * (psiex(l, kII * a, x) + psi2ex(l, kII * a, x)) * MuII
                                              + mDR[jj] * (xiex(l, kI * a, x) + xi2ex(l, kI * a, x)) * MuI)                                                          # magnetic parameter3 right

    # define points for least square fitting [similar to operating on
    # equations with int_ {-1}**(+1} dx legendrePl(m,x)...]
    x = {}
    for m in range(1, int(mmax / 3) + 1):
        x[int(m)] = (-1 + 0.1 * (m - 1) / (mmax / 3))
        x[int(m + mmax / 3)] = (-0.9 + 1.8 * m / (mmax / 3))
        x[int(m + 2 * mmax / 3)] = (+0.9 + 0.1 * m / (mmax / 3))

    efun1aL = np.zeros((mmax, lmax), dtype=complex)
    efun1bL = np.zeros((mmax, lmax), dtype=complex)
    efun2aL = np.zeros((mmax, lmax), dtype=complex)
    efun2bL = np.zeros((mmax, lmax), dtype=complex)
    efun3L = np.zeros((mmax, lmax), dtype=complex)

    mfun1aL = np.zeros((mmax, lmax), dtype=complex)
    mfun1bL = np.zeros((mmax, lmax), dtype=complex)
    mfun2aL = np.zeros((mmax, lmax), dtype=complex)
    mfun2bL = np.zeros((mmax, lmax), dtype=complex)
    mfun3L = np.zeros((mmax, lmax), dtype=complex)

    efun1aR = np.zeros((mmax, lmax), dtype=complex)
    efun1bR = np.zeros((mmax, lmax), dtype=complex)
    efun2aR = np.zeros((mmax, lmax), dtype=complex)
    efun2bR = np.zeros((mmax, lmax), dtype=complex)
    efun3R = np.zeros((mmax, lmax), dtype=complex)

    mfun1aR = np.zeros((mmax, lmax), dtype=complex)
    mfun1bR = np.zeros((mmax, lmax), dtype=complex)
    mfun2aR = np.zeros((mmax, lmax), dtype=complex)
    mfun2bR = np.zeros((mmax, lmax), dtype=complex)
    mfun3R = np.zeros((mmax, lmax), dtype=complex)

    for ii in range(mmax):
        m = ii + 1
        # define points for least square fitting [similar to operating on equations with int_ {-1}**(+1} dx legendrePl(m,x)...]
        # the points x[m] lie in the range [-1,1] or similarly th(m) is in the range [0,pi]
        # x[m] = (-1 + 2*(m-1)/(mmax-1))
        for jj in range(lmax):
            l = jj + 1
            # left
            efun1aL[ii, jj] = eLambda1L[l](x[m]) * legendrePl(l, x[m])
            efun1bL[ii, jj] = eLambda1L[l](x[m]) * legendrePlex1(l, x[m])
            mfun1aL[ii, jj] = mLambda1L[l](x[m]) * legendrePl(l, x[m])
            mfun1bL[ii, jj] = mLambda1L[l](x[m]) * legendrePlex1(l, x[m])
            mfun2aL[ii, jj] = mLambda2L[l](x[m]) * legendrePl(l, x[m])
            mfun2bL[ii, jj] = mLambda2L[l](x[m]) * legendrePlex1(l, x[m])
            efun2aL[ii, jj] = eLambda2L[l](x[m]) * legendrePl(l, x[m])
            efun2bL[ii, jj] = eLambda2L[l](x[m]) * legendrePlex1(l, x[m])
            efun3L[ii, jj] = eLambda3L[l](x[m]) * legendrePl(l, x[m])
            mfun3L[ii, jj] = mLambda3L[l](x[m]) * legendrePl(l, x[m])

            # right
            efun1aR[ii, jj] = eLambda1R[l](x[m]) * legendrePl(l, x[m])
            efun1bR[ii, jj] = eLambda1R[l](x[m]) * legendrePlex1(l, x[m])
            mfun1aR[ii, jj] = mLambda1R[l](x[m]) * legendrePl(l, x[m])
            mfun1bR[ii, jj] = mLambda1R[l](x[m]) * legendrePlex1(l, x[m])
            mfun2aR[ii, jj] = mLambda2R[l](x[m]) * legendrePl(l, x[m])
            mfun2bR[ii, jj] = mLambda2R[l](x[m]) * legendrePlex1(l, x[m])
            efun2aR[ii, jj] = eLambda2R[l](x[m]) * legendrePl(l, x[m])
            efun2bR[ii, jj] = eLambda2R[l](x[m]) * legendrePlex1(l, x[m])
            efun3R[ii, jj] = eLambda3R[l](x[m]) * legendrePl(l, x[m])
            mfun3R[ii, jj] = mLambda3R[l](x[m]) * legendrePl(l, x[m])

    # first order BC can be written in form: M11*eCc + M12*eDd + M13*mCc + M14*mDd = N1
    #                                        ...
    #                                        M61*eCc + M62*eDd + M63*mCc + M64*mDd = N6
    # M11...M66 are matrices including the pre-factors for the individual terms in the sums of eCc[jj], etc
    # eCc...mDd are vectors including all the expansion coefficients (eCc = [eCc(1),...,eCc(lmax)])
    # N1...N6 include the 0-th order terms and angular-dependent Legendre and
    # Bessel functions

    N1L = np.zeros(mmax, dtype=complex)
    N2L = np.zeros(mmax, dtype=complex)
    N3L = np.zeros(mmax, dtype=complex)
    N4L = np.zeros(mmax, dtype=complex)
    N5L = np.zeros(mmax, dtype=complex)
    N6L = np.zeros(mmax, dtype=complex)

    N1R = np.zeros(mmax, dtype=complex)
    N2R = np.zeros(mmax, dtype=complex)
    N3R = np.zeros(mmax, dtype=complex)
    N4R = np.zeros(mmax, dtype=complex)
    N5R = np.zeros(mmax, dtype=complex)
    N6R = np.zeros(mmax, dtype=complex)

    for ii in range(mmax):
        # left
        N1L[ii] = np.sum(efun1aL[ii, :]) - np.sum(mfun2bL[ii, :])
        N2L[ii] = np.sum(efun1bL[ii, :]) - np.sum(mfun2aL[ii, :])
        N3L[ii] = np.sum(mfun1aL[ii, :]) - np.sum(efun2bL[ii, :])
        N4L[ii] = np.sum(mfun1bL[ii, :]) - np.sum(efun2aL[ii, :])
        N5L[ii] = np.sum(efun3L[ii, :])
        N6L[ii] = np.sum(mfun3L[ii, :])
        # right
        N1R[ii] = np.sum(efun1aR[ii, :]) - np.sum(mfun2bR[ii, :])
        N2R[ii] = np.sum(efun1bR[ii, :]) - np.sum(mfun2aR[ii, :])
        N3R[ii] = np.sum(mfun1aR[ii, :]) - np.sum(efun2bR[ii, :])
        N4R[ii] = np.sum(mfun1bR[ii, :]) - np.sum(efun2aR[ii, :])
        N5R[ii] = np.sum(efun3R[ii, :])
        N6R[ii] = np.sum(mfun3R[ii, :])

    ##
    M11 = np.zeros((mmax, lmax), dtype=complex)
    M12 = np.zeros((mmax, lmax), dtype=complex)
    M13 = np.zeros((mmax, lmax), dtype=complex)
    M14 = np.zeros((mmax, lmax), dtype=complex)

    M21 = np.zeros((mmax, lmax), dtype=complex)
    M22 = np.zeros((mmax, lmax), dtype=complex)
    M23 = np.zeros((mmax, lmax), dtype=complex)
    M24 = np.zeros((mmax, lmax), dtype=complex)

    M31 = np.zeros((mmax, lmax), dtype=complex)
    M32 = np.zeros((mmax, lmax), dtype=complex)
    M33 = np.zeros((mmax, lmax), dtype=complex)
    M34 = np.zeros((mmax, lmax), dtype=complex)

    M41 = np.zeros((mmax, lmax), dtype=complex)
    M42 = np.zeros((mmax, lmax), dtype=complex)
    M43 = np.zeros((mmax, lmax), dtype=complex)
    M44 = np.zeros((mmax, lmax), dtype=complex)

    M51 = np.zeros((mmax, lmax), dtype=complex)
    M52 = np.zeros((mmax, lmax), dtype=complex)
    M53 = np.zeros((mmax, lmax), dtype=complex)
    M54 = np.zeros((mmax, lmax), dtype=complex)

    M61 = np.zeros((mmax, lmax), dtype=complex)
    M62 = np.zeros((mmax, lmax), dtype=complex)
    M63 = np.zeros((mmax, lmax), dtype=complex)
    M64 = np.zeros((mmax, lmax), dtype=complex)

    for ii in range(mmax):
        m = ii + 1
        for jj in range(lmax):
            l = jj + 1

            M11[ii, jj] = +1 / kII * \
                psi1xx(l, kII * a, x[m]) * legendrePl(l, x[m])
            M12[ii, jj] = -1 / kI * \
                xi1xx(l, kI * a, x[m]) * legendrePl(l, x[m])
            M13[ii, jj] = +1 / k1II * \
                psixx(l, kII * a, x[m]) * legendrePlex1(l, x[m])
            M14[ii, jj] = -1 / k1I * \
                xixx(l, kI * a, x[m]) * legendrePlex1(l, x[m])

            M21[ii, jj] = +1 / kII * \
                psi1xx(l, kII * a, x[m]) * legendrePlex1(l, x[m])
            M22[ii, jj] = -1 / kI * \
                xi1xx(l, kI * a, x[m]) * legendrePlex1(l, x[m])
            M23[ii, jj] = +1 / k1II * \
                psixx(l, kII * a, x[m]) * legendrePl(l, x[m])
            M24[ii, jj] = -1 / k1I * \
                xixx(l, kI * a, x[m]) * legendrePl(l, x[m])

            M31[ii, jj] = +1 / k2II * \
                psixx(l, kII * a, x[m]) * legendrePlex1(l, x[m])
            M32[ii, jj] = -1 / k2I * \
                xixx(l, kI * a, x[m]) * legendrePlex1(l, x[m])
            M33[ii, jj] = M11[ii, jj]
            M34[ii, jj] = M12[ii, jj]

            M41[ii, jj] = +1 / k2II * \
                psixx(l, kII * a, x[m]) * legendrePl(l, x[m])
            M42[ii, jj] = -1 / k2I * \
                xixx(l, kI * a, x[m]) * legendrePl(l, x[m])
            M43[ii, jj] = M21[ii, jj]
            M44[ii, jj] = M22[ii, jj]

            M51[ii, jj] = + k1II * (psixx(l, kII * a, x[m]) +
                                    psi2xx(l, kII * a, x[m])) * legendrePl(l, x[m])
            M52[ii, jj] = - k1I * (xixx(l, kI * a, x[m]) +
                                   xi2xx(l, kI * a, x[m])) * legendrePl(l, x[m])
            M53[ii, jj] = 0
            M54[ii, jj] = 0

            M61[ii, jj] = 0
            M62[ii, jj] = 0
            M63[ii, jj] = + MuII * (psixx(l, kII * a, x[m]) +
                                    psi2xx(l, kII * a, x[m])) * legendrePl(l, x[m])
            M64[ii, jj] = - MuI * (xixx(l, kI * a, x[m]) +
                                   xi2xx(l, kI * a, x[m])) * legendrePl(l, x[m])

    Matrix = np.zeros((6 * mmax, 6 * lmax), dtype=complex)

    for ii in range(mmax):
        m = ii + 1
        for jj in range(lmax):
            l = jj + 1
            Matrix[ii, jj] = M11[ii, jj]
            Matrix[ii, jj + lmax] = M12[ii, jj]
            Matrix[ii, jj + 2 * lmax] = M13[ii, jj]
            Matrix[ii, jj + 3 * lmax] = M14[ii, jj]

            Matrix[ii + mmax, jj] = M21[ii, jj]
            Matrix[ii + mmax, jj + lmax] = M22[ii, jj]
            Matrix[ii + mmax, jj + 2 * lmax] = M23[ii, jj]
            Matrix[ii + mmax, jj + 3 * lmax] = M24[ii, jj]

            Matrix[ii + 2 * mmax, jj] = M31[ii, jj]
            Matrix[ii + 2 * mmax, jj + lmax] = M32[ii, jj]
            Matrix[ii + 2 * mmax, jj + 2 * lmax] = M33[ii, jj]
            Matrix[ii + 2 * mmax, jj + 3 * lmax] = M34[ii, jj]

            Matrix[ii + 3 * mmax, jj] = M41[ii, jj]
            Matrix[ii + 3 * mmax, jj + lmax] = M42[ii, jj]
            Matrix[ii + 3 * mmax, jj + 2 * lmax] = M43[ii, jj]
            Matrix[ii + 3 * mmax, jj + 3 * lmax] = M44[ii, jj]

            Matrix[ii + 4 * mmax, jj] = M51[ii, jj]
            Matrix[ii + 4 * mmax, jj + lmax] = M52[ii, jj]
            Matrix[ii + 4 * mmax, jj + 2 * lmax] = M53[ii, jj]
            Matrix[ii + 4 * mmax, jj + 3 * lmax] = M54[ii, jj]

            Matrix[ii + 5 * mmax, jj] = M61[ii, jj]
            Matrix[ii + 5 * mmax, jj + lmax] = M62[ii, jj]
            Matrix[ii + 5 * mmax, jj + 2 * lmax] = M63[ii, jj]
            Matrix[ii + 5 * mmax, jj + 3 * lmax] = M64[ii, jj]

    VectorL = np.zeros(6 * mmax, dtype=complex)
    VectorR = np.zeros(6 * mmax, dtype=complex)
    for ii in range(mmax):
        # left and right
        VectorL[ii] = N1L[ii]
        VectorL[ii + mmax] = N2L[ii]
        VectorL[ii + 2 * mmax] = N3L[ii]
        VectorL[ii + 3 * mmax] = N4L[ii]
        VectorL[ii + 4 * mmax] = N5L[ii]
        VectorL[ii + 5 * mmax] = N6L[ii]

        VectorR[ii] = N1R[ii]
        VectorR[ii + mmax] = N2R[ii]
        VectorR[ii + 2 * mmax] = N3R[ii]
        VectorR[ii + 3 * mmax] = N4R[ii]
        VectorR[ii + 4 * mmax] = N5R[ii]
        VectorR[ii + 5 * mmax] = N6R[ii]

    Weight = np.zeros(6 * mmax, dtype=complex)
    # weights
    for ii in range(mmax):
        m = ii + 1
        Weight[ii] = 1
        Weight[ii + mmax] = 1
        Weight[ii + 2 * mmax] = 1
        Weight[ii + 3 * mmax] = 1
        Weight[ii + 4 * mmax] = 1 / (100 * m**0.8)
        Weight[ii + 5 * mmax] = 1 / (22 * m**0.20)

    # Weights were commented out?:
    # xL = lscov (Matrix,VectorL.',Weight.').'
    xL = lscov(np.matrix(Matrix), np.matrix(VectorL).T, np.matrix(Weight).T)
    xR = lscov(np.matrix(Matrix), np.matrix(VectorR).T, np.matrix(Weight).T)

    if verbose:
        print('If Eps=0, ignore previous error messages regarding rank deficiency!')
    eCcL = np.zeros(lmax, dtype=complex)
    eDdL = np.zeros(lmax, dtype=complex)
    mCcL = np.zeros(lmax, dtype=complex)
    mDdL = np.zeros(lmax, dtype=complex)

    eCcR = np.zeros(lmax, dtype=complex)
    eDdR = np.zeros(lmax, dtype=complex)
    mCcR = np.zeros(lmax, dtype=complex)
    mDdR = np.zeros(lmax, dtype=complex)

    for jj in range(lmax):
        l = jj + 1
        # left and right first order expansion coefficients
        eCcL[jj] = xL[jj]
        eDdL[jj] = xL[jj + lmax]
        mCcL[jj] = xL[jj + 2 * lmax]
        mDdL[jj] = xL[jj + 3 * lmax]

        eCcR[jj] = xR[jj] * (-1)**(l - 1)
        eDdR[jj] = xR[jj + lmax] * (-1)**(l - 1)
        mCcR[jj] = xR[jj + 2 * lmax] * (-1)**l
        mDdR[jj] = xR[jj + 3 * lmax] * (-1)**l

        # corrected expansion coefficients for right optical fiber*
        eBR[jj] = eBR[jj] * (-1)**(l - 1)
        eCR[jj] = eCR[jj] * (-1)**(l - 1)
        eDR[jj] = eDR[jj] * (-1)**(l - 1)
        mBR[jj] = mBR[jj] * (-1)**l
        mCR[jj] = mCR[jj] * (-1)**l
        mDR[jj] = mDR[jj] * (-1)**l
        #*additional factors [(-1)**(l-1),(-1)**l] should have been included when eB, mB are calculated however due to least squares approach slight antisymmetry arises...
        # hence include factors only after least square approach

    # Field Expansions For Incident Fields
    # sums of radial, zenithal and azimuthal field terms from l=1 to l=lmax
    # initialisation of sums
    # Paul: These are not used?!
    #EriL   = lambda r, th, phi: 0
    #EriR   = lambda r, th, phi: 0
    #EthiL  = lambda r, th, phi: 0
    #EthiR  = lambda r, th, phi: 0
    #EphiiL = lambda r, th, phi: 0
    #EphiiR = lambda r, th, phi: 0
    #HriL   = lambda r, th, phi: 0
    #HriR   = lambda r, th, phi: 0
    #HthiL  = lambda r, th, phi: 0
    #HthiR  = lambda r, th, phi: 0
    #HphiiL = lambda r, th, phi: 0
    #HphiiR = lambda r, th, phi: 0
    # left
    #EriL   = lambda r, th, phi: EriL(r,th,phi)   + eBL[l-1]*(psi(l,kI*r)+psi2(l,kI*r))*legendrePl(l,np.cos(th))*np.cos(phi)
    # EthiL  = lambda r, th, phi: EthiL(r,th,phi)  -(eBL[l-1]*psi1 (l,kI*r)/(kI*r)*legendrePl1(l,np.cos(th))*np.sin(th) \
    #                                        + mBL[l-1]*psi (l,kI*r)/(k1I*r)*legendrePl (l,np.cos(th))/np.sin(th))*np.cos(phi)
    # EphiiL = lambda r, th, phi: EphiiL(r,th,phi) -(eBL[l-1]*psi1 (l,kI*r)/(kI*r)*legendrePl (l,np.cos(th))/np.sin(th) \
    #                                        + mBL[l-1]*psi (l,kI*r)/(k1I*r)*legendrePl1(l,np.cos(th))*np.sin(th))*np.sin(phi)
    #HriL   = lambda r, th, phi: HriL(r,th,phi)   + mBL[l-1]*(psi(l,kI*r)+psi2(l,kI*r))*legendrePl(l,np.cos(th))*np.sin(phi)
    # HthiL  = lambda r, th, phi: HthiL(r,th,phi)  -(mBL[l-1]*psi1 (l,kI*r)/(kI*r)*legendrePl1(l,np.cos(th))*np.sin(th) \
    #                                        + eBL[l-1]*psi (l,kI*r)/(k2I*r)*legendrePl (l,np.cos(th))/np.sin(th))*np.sin(phi)
    # HphiiL = lambda r, th, phi: HphiiL(r,th,phi) +(mBL[l-1]*psi1 (l,kI*r)/(kI*r)*legendrePl (l,np.cos(th))/np.sin(th) \
    #                                        + eBL[l-1]*psi (l,kI*r)/(k2I*r)*legendrePl1(l,np.cos(th))*np.sin(th))*np.cos(phi)
    # right
    #EriR   = lambda r, th, phi: EriR(r,th,phi)   + eBR[l-1]*(psi(l,kI*r)+psi2(l,kI*r))*legendrePl(l,np.cos(th))*np.cos(phi)
    # EthiR  = lambda r, th, phi: EthiR(r,th,phi)  -(eBR[l-1]*psi1 (l,kI*r)/(kI*r)*legendrePl1(l,np.cos(th))*np.sin(th) \
    #                                        + mBR[l-1]*psi (l,kI*r)/(k1I*r)*legendrePl (l,np.cos(th))/np.sin(th))*np.cos(phi)
    # EphiiR = lambda r, th, phi: EphiiR(r,th,phi) -(eBR[l-1]*psi1 (l,kI*r)/(kI*r)*legendrePl (l,np.cos(th))/np.sin(th) \
    #                                        + mBR[l-1]*psi (l,kI*r)/(k1I*r)*legendrePl1(l,np.cos(th))*np.sin(th))*np.sin(phi)
    #HriR   = lambda r, th, phi: HriR(r,th,phi)   + mBR[l-1]*(psi(l,kI*r)+psi2(l,kI*r))*legendrePl(l,np.cos(th))*np.sin(phi)
    # HthiR  = lambda r, th, phi: HthiR(r,th,phi)  -(mBR[l-1]*psi1 (l,kI*r)/(kI*r)*legendrePl1(l,np.cos(th))*np.sin(th) \
    #                                        + eBR[l-1]*psi (l,kI*r)/(k2I*r)*legendrePl (l,np.cos(th))/np.sin(th))*np.sin(phi)
    # HphiiR = lambda r, th, phi: HphiiR(r,th,phi) +(mBR[l-1]*psi1 (l,kI*r)/(kI*r)*legendrePl (l,np.cos(th))/np.sin(th) \
    #                                        + eBR[l-1]*psi (l,kI*r)/(k2I*r)*legendrePl1(l,np.cos(th))*np.sin(th))*np.cos(phi)

    # Field Expansions For Scattered Fields
    # sums of radial, zenithal and azimuthal field terms from l=1 to l=lmax
    # Paul: workaround to recursing into lambda functions
    def wrapper_expansion(lambda_func):
        def wrapped(r, th, ph):
            result = 0
            for jj in range(lmax):
                l = jj + 1
                result += lambda_func(r, th, ph, l)
            return result
        return wrapped
    # left

    def ErsL_it(r, th, phi, l): return (
        eDL[l - 1] + eDdL[l - 1]) * (xi(l, kI * r) + xi2(l, kI * r)) * legendrePl(l, np.cos(th)) * np.cos(phi)

    def EthsL_it(r, th, phi, l): return -((eDL[l - 1] + eDdL[l - 1]) * xi1(l, kI * r) / (kI * r) * legendrePl1(l, np.cos(th)) * np.sin(th)
                                          + (mDL[l - 1] + mDdL[l - 1]) * xi(l, kI * r) / (k1I * r) * legendrePl(l, np.cos(th)) / np.sin(th)) * np.cos(phi)

    def EphisL_it(r, th, phi, l): return -((eDL[l - 1] + eDdL[l - 1]) * xi1(l, kI * r) / (kI * r) * legendrePl(l, np.cos(th)) / np.sin(th)
                                           + (mDL[l - 1] + mDdL[l - 1]) * xi(l, kI * r) / (k1I * r) * legendrePl1(l, np.cos(th)) * np.sin(th)) * np.sin(phi)

    def HrsL_it(r, th, phi, l): return +(mDL[l - 1] + mDdL[l - 1]) * (
        xi(l, kI * r) + xi2(l, kI * r)) * legendrePl(l, np.cos(th)) * np.sin(phi)

    def HthsL_it(r, th, phi, l): return -((mDL[l - 1] + mDdL[l - 1]) * xi1(l, kI * r) / (kI * r) * legendrePl1(l, np.cos(th)) * np.sin(th)
                                          + (eDL[l - 1] + eDdL[l - 1]) * xi(l, kI * r) / (k2I * r) * legendrePl(l, np.cos(th)) / np.sin(th)) * np.sin(phi)

    def HphisL_it(r, th, phi, l): return +((mDL[l - 1] + mDdL[l - 1]) * xi1(l, kI * r) / (kI * r) * legendrePl(l, np.cos(th)) / np.sin(th)
                                           + (eDL[l - 1] + eDdL[l - 1]) * xi(l, kI * r) / (k2I * r) * legendrePl1(l, np.cos(th)) * np.sin(th)) * np.cos(phi)
    # right

    def ErsR_it(r, th, phi, l): return +(eDR[l - 1] + eDdR[l - 1]) * (
        xi(l, kI * r) + xi2(l, kI * r)) * legendrePl(l, np.cos(th)) * np.cos(phi)

    def EthsR_it(r, th, phi, l): return -((eDR[l - 1] + eDdR[l - 1]) * xi1(l, kI * r) / (kI * r) * legendrePl1(l, np.cos(th)) * np.sin(th)
                                          + (mDR[l - 1] + mDdR[l - 1]) * xi(l, kI * r) / (k1I * r) * legendrePl(l, np.cos(th)) / np.sin(th)) * np.cos(phi)

    def EphisR_it(r, th, phi, l): return -((eDR[l - 1] + eDdR[l - 1]) * xi1(l, kI * r) / (kI * r) * legendrePl(l, np.cos(th)) / np.sin(th)
                                           + (mDR[l - 1] + mDdR[l - 1]) * xi(l, kI * r) / (k1I * r) * legendrePl1(l, np.cos(th)) * np.sin(th)) * np.sin(phi)

    def HrsR_it(r, th, phi, l): return +(mDR[l - 1] + mDdR[l - 1]) * (
        xi(l, kI * r) + xi2(l, kI * r)) * legendrePl(l, np.cos(th)) * np.sin(phi)

    def HthsR_it(r, th, phi, l): return -((mDR[l - 1] + mDdR[l - 1]) * xi1(l, kI * r) / (kI * r) * legendrePl1(l, np.cos(th)) * np.sin(th)
                                          + (eDR[l - 1] + eDdR[l - 1]) * xi(l, kI * r) / (k2I * r) * legendrePl(l, np.cos(th)) / np.sin(th)) * np.sin(phi)

    def HphisR_it(r, th, phi, l): return +((mDR[l - 1] + mDdR[l - 1]) * xi1(l, kI * r) / (kI * r) * legendrePl(l, np.cos(th)) / np.sin(th)
                                           + (eDR[l - 1] + eDdR[l - 1]) * xi(l, kI * r) / (k2I * r) * legendrePl1(l, np.cos(th)) * np.sin(th)) * np.cos(phi)

    ErsL = wrapper_expansion(ErsL_it)
    EthsL = wrapper_expansion(EthsL_it)
    EphisL = wrapper_expansion(EphisL_it)
    HrsL = wrapper_expansion(HrsL_it)
    HthsL = wrapper_expansion(HthsL_it)
    HphisL = wrapper_expansion(HphisL_it)

    ErsR = wrapper_expansion(ErsR_it)
    EthsR = wrapper_expansion(EthsR_it)
    EphisR = wrapper_expansion(EphisR_it)
    HrsR = wrapper_expansion(HrsR_it)
    HthsR = wrapper_expansion(HthsR_it)
    HphisR = wrapper_expansion(HphisR_it)

    # Field Expansions For Internal Fields (Within)
    # sums of radial, zenithal and azimuthal field terms from l=1 to l=lmax
    # initialisation of sums
    def ErwL_it(r, th, phi, l): return +(eCL[l - 1] + eCcL[l - 1]) * (
        psi(l, kII * r) + psi2(l, kII * r)) * legendrePl(l, np.cos(th)) * np.cos(phi)

    def EthwL_it(r, th, phi, l): return -((eCL[l - 1] + eCcL[l - 1]) * psi1(l, kII * r) / (kII * r) * legendrePl1(l, np.cos(th)) * np.sin(th)
                                          + (mCL[l - 1] + mCcL[l - 1]) * psi(l, kII * r) / (k1II * r) * legendrePl(l, np.cos(th)) / np.sin(th)) * np.cos(phi)

    def EphiwL_it(r, th, phi, l): return -((eCL[l - 1] + eCcL[l - 1]) * psi1(l, kII * r) / (kII * r) * legendrePl(l, np.cos(th)) / np.sin(th)
                                           + (mCL[l - 1] + mCcL[l - 1]) * psi(l, kII * r) / (k1II * r) * legendrePl1(l, np.cos(th)) * np.sin(th)) * np.sin(phi)
    # Paul: These are not used?!
    #HrwL_it   = lambda r,th,phi,l: +(mCL[l-1] + mCcL[l-1])*(psi(l,kII*r)+psi2(l,kII*r))*legendrePl(l,np.cos(th))*np.sin(phi)
    # HthwL_it  = lambda r,th,phi,l: -((mCL[l-1] + mCcL[l-1])*psi1 (l,kII*r)/(kII*r)*legendrePl1(l,np.cos(th))*np.sin(th) \
    #                              + (eCL[l-1] + eCcL[l-1])*psi (l,kII*r)/(k2II*r)*legendrePl (l,np.cos(th))/np.sin(th))*np.sin(phi)
    # HphiwL_it = lambda r,th,phi,l: +((mCL[l-1] + mCcL[l-1])*psi1 (l,kII*r)/(kII*r)*legendrePl (l,np.cos(th))/np.sin(th) \
    #                              + (eCL[l-1] + eCcL[l-1])*psi (l,kII*r)/(k2II*r)*legendrePl1(l,np.cos(th))*np.sin(th))*np.cos(phi)
    # right
    #ErwR_it   = lambda r,th,phi,l: +(eCR[l-1] + eCcR[l-1])*(psi(l,kII*r)+psi2(l,kII*r))*legendrePl(l,np.cos(th))*np.cos(phi)
    # EthwR_it  = lambda r,th,phi,l: -((eCR[l-1] + eCcR[l-1])*psi1 (l,kII*r)/(kII*r)*legendrePl1(l,np.cos(th))*np.sin(th) \
    #                              + (mCR[l-1] + mCcR[l-1])*psi (l,kII*r)/(k1II*r)*legendrePl (l,np.cos(th))/np.sin(th))*np.cos(phi)
    # EphiwR_it = lambda r,th,phi,l: -((eCR[l-1] + eCcR[l-1])*psi1 (l,kII*r)/(kII*r)*legendrePl (l,np.cos(th))/np.sin(th) \
    #                              + (mCR[l-1] + mCcR[l-1])*psi (l,kII*r)/(k1II*r)*legendrePl1(l,np.cos(th))*np.sin(th))*np.sin(phi)
    #HrwR_it   = lambda r,th,phi,l: +(mCR[l-1] + mCcR[l-1])*(psi(l,kII*r)+psi2(l,kII*r))*legendrePl(l,np.cos(th))*np.sin(phi)
    # HthwR_it  = lambda r,th,phi,l: -((mCR[l-1] + mCcR[l-1])*psi1 (l,kII*r)/(kII*r)*legendrePl1(l,np.cos(th))*np.sin(th) \
    #                              + (eCR[l-1] + eCcR[l-1])*psi (l,kII*r)/(k2II*r)*legendrePl (l,np.cos(th))/np.sin(th))*np.sin(phi)
    # HphiwR_it = lambda r,th,phi,l: +((mCR[l-1] + mCcR[l-1])*psi1 (l,kII*r)/(kII*r)*legendrePl (l,np.cos(th))/np.sin(th) \
    #                              + (eCR[l-1] + eCcR[l-1])*psi (l,kII*r)/(k2II*r)*legendrePl1(l,np.cos(th))*np.sin(th))*np.cos(phi)

    ErwL = wrapper_expansion(ErwL_it)
    EthwL = wrapper_expansion(EthwL_it)
    EphiwL = wrapper_expansion(EphiwL_it)
    # Paul: These are not used?!
    #HrwL = wrapper_expansion(HrwL_it)
    #HthwL = wrapper_expansion(HthwL_it)
    #HphiwL = wrapper_expansion(HphiwL_it)
    #
    #ErwR = wrapper_expansion(ErwR_it)
    #EthwR = wrapper_expansion(EthwR_it)
    #EphiwR = wrapper_expansion(EphiwR_it)
    #HrwR = wrapper_expansion(HrwR_it)
    #HthwR = wrapper_expansion(HthwR_it)
    #HphiwR = wrapper_expansion(HphiwR_it)

    # Overall Fields
    # sum of the electric fields in region I (independent of time) [separated
    # for left and right]
    def ErL(r, th, phi): return ErsL(r, th, phi) + eEriL(r, th, phi)

    def ErR(r, th, phi): return ErsR(r, th, phi) + eEriR(r, th, phi)

    def EthL(r, th, phi): return EthsL(r, th, phi) + eEthiL(r, th, phi)

    def EthR(r, th, phi): return EthsR(r, th, phi) + eEthiR(r, th, phi)

    def EphiL(r, th, phi): return EphisL(r, th, phi) + eEphiiL(r, th, phi)

    def EphiR(r, th, phi): return EphisR(r, th, phi) + eEphiiR(r, th, phi)

    def HrL(r, th, phi): return HrsL(r, th, phi) + eHriL(r, th, phi)

    def HrR(r, th, phi): return HrsR(r, th, phi) + eHriR(r, th, phi)

    def HthL(r, th, phi): return HthsL(r, th, phi) + eHthiL(r, th, phi)

    def HthR(r, th, phi): return HthsR(r, th, phi) + eHthiR(r, th, phi)

    def HphiL(r, th, phi): return HphisL(r, th, phi) + eHphiiL(r, th, phi)

    def HphiR(r, th, phi): return HphisR(r, th, phi) + eHphiiR(r, th, phi)

    # Checking Boundary Conditions
    # generate points on boundary
    r = np.zeros(22, dtype=complex)
    th = np.zeros(22, dtype=complex)
    phi = np.pi / 4
    for ii in range(22):
        th[ii] = np.pi / 20.003 * (ii + 1 - 0.999)
        r[ii] = a * B1(np.cos(th[ii]))

    FEr1 = np.zeros(22, dtype=float)
    FEr2 = np.zeros(22, dtype=float)
    FEth1 = np.zeros(22, dtype=float)
    FEth2 = np.zeros(22, dtype=float)
    FEphi1 = np.zeros(22, dtype=float)
    FEphi2 = np.zeros(22, dtype=float)

    FEr = np.zeros(22, dtype=float)
    FEth = np.zeros(22, dtype=float)
    FEphi = np.zeros(22, dtype=float)

    for ii in range(22):
        # internal and external fields
        FEr1[ii] = np.real(EpsilonI * ErL(r[ii], th[ii], phi))
        FEr2[ii] = np.real(EpsilonII * ErwL(r[ii], th[ii], phi))
        FEth1[ii] = np.real(EthL(r[ii], th[ii], phi))
        FEth2[ii] = np.real(EthwL(r[ii], th[ii], phi))
        FEphi1[ii] = np.real(EphiL(r[ii], th[ii], phi))
        FEphi2[ii] = np.real(EphiwL(r[ii], th[ii], phi))
        # ratios
        FEr[ii] = FEr1[ii] / FEr2[ii]
        FEth[ii] = FEth1[ii] / FEth2[ii]
        FEphi[ii] = FEphi1[ii] / FEphi2[ii]

    if verbose:
        print('Following output indicates the accuracy of the matching of the boundary conditions.')
        print('theta, Er, Er_ratio, Eth, Eth_ratio, Ephi, Ephi_ratio.')
        print(np.array([th, FEr1, FEr, FEth1, FEth, FEphi1, FEphi]))
        print('The ratio should be close to 1 for good matching of the BCs. However if the fields tend to zero the ratio is not that relevant.')

    # Data for Plotting
    # time-averaged, radial stress on infinitesimal area dA = r**2 np.sin(th)
    # dr dth dphi
    def sigmarr(r, th, phi): return -1 / (16 * np.pi) * np.real(
        EpsilonI * (ErL(r, th, phi) * np.conj(ErL(r, th, phi)) - EthL(r, th, phi) *
                    np.conj(EthL(r, th, phi)) - EphiL(r, th, phi) * np.conj(EphiL(r, th, phi)))
        + EpsilonI * (ErR(r, th, phi) * np.conj(ErR(r, th, phi)) - EthR(r, th, phi) *
                      np.conj(EthR(r, th, phi)) - EphiR(r, th, phi) * np.conj(EphiR(r, th, phi)))
        + MuI * (HrL(r, th, phi) * np.conj(HrL(r, th, phi)) - HthL(r, th, phi) *
                 np.conj(HthL(r, th, phi)) - HphiL(r, th, phi) * np.conj(HphiL(r, th, phi)))
        + MuI * (HrR(r, th, phi) * np.conj(HrR(r, th, phi)) - HthR(r, th, phi) * np.conj(HthR(r, th, phi)) - HphiR(r, th, phi) * np.conj(HphiR(r, th, phi))))

    th = np.zeros(n_points, dtype=float)
    r = np.zeros(n_points, dtype=float)
    sigma = np.zeros(n_points, dtype=float)

    for ii in range(n_points):
        th[ii] = theta_max / (n_points - 0.998) * (ii + 1 - 0.999)
        r[ii] = a * B1(np.cos(th[ii]))
        sigma[ii] = sigmarr(r[ii], th[ii], np.pi / 4)

    res = [th, sigma]

    if ret_legendre_decomp:
        coeff = stress2legendre(stress=sigma, theta=th, n_poly=lmax)
        res.append(coeff)

    return res



"""Look-up table for GGF computation"""
import pathlib
from pkg_resources import resource_filename

import h5py
import numpy as np
from scipy.interpolate import interpn


class NotInLUTError(BaseException):
    pass


def _match_lut(kwargs, lut_path=None):
    """Find matching LUT for a set of keyword arguments

    Parameters
    ----------
    kwargs: dict
        keyword arguments for ggf.get_ggf
    lut_path: str, pathlib.Path or None
        path to an hdf5 file containing a LUT; if None,
        the default LUTs will be searched

    Returns
    -------
    lut_path: pathlib.Path
        path to matching LUT hdf5 file
    """
    if lut_path:
        paths = [pathlib.Path(lut_path)]
    else:
        paths = get_lut_paths()
    for path in paths:
        with h5py.File(path, mode="r") as h5:
            for key in kwargs:
                if _param_in_lut(key, kwargs[key], h5["lut"].attrs):
                    pass  # everything OK so far
                else:
                    break  # try next LUT
            else:
                break  # this is the right LUT
    else:
        msg = "No matching LUT found for: {} in {}".format(kwargs, paths)
        raise NotInLUTError(msg)
    return path


def _param_in_lut(key, value, h5_attrs):
    """Test whether the given parameter is accessible via a LUT

    Parameters
    ----------
    key: str
        keyword name (see ggf.get_ggf)
    value: str or float
        keyword value to check
    h5_attrs: dict
        attributes from a LUT hdf5 file

    Returns
    -------
    inlut: bool
        True, if the key-value pair is covered by the LUT
    """
    inlut = False
    if key == "n_poly" and value is None:  # default value is given in LUT
        inlut = True
    elif key in h5_attrs:  # fixed parameter
        if h5_attrs[key] == value or value is None:
            inlut = True
    elif isinstance(value, str):  # model
        inlut = False
    else:  # range
        vmin = h5_attrs["{} min".format(key)]
        vmax = h5_attrs["{} max".format(key)]
        if value <= vmax and value >= vmin:
            inlut = True
        else:
            inlut = False
    return inlut


def get_lut_paths():
    """Return a list of look-up table hdf5 files in ggf"""
    lutpath = pathlib.Path(resource_filename("ggf", "lut"))
    paths = lutpath.glob("*.h5")
    return sorted(paths)


def get_ggf_lut(model, semi_major, semi_minor, object_index, medium_index,
                effective_fiber_distance, mode_field_diameter,
                power_per_fiber, wavelength, poisson_ratio,
                n_poly=None, lut_path=None, verbose=False):
    """Linear interpolation of the GGF from a look-up table

    Parameters
    ----------
    model: str
        Model to use, one of: `boyde2009`
    semi_major: float
        Semi-major axis of an ellipse fit to the object perimeter [m]
    semi_minor: float
        Semi-minor axis of an ellipse fit to the object perimeter [m]
    object_index: float
        Refractive index of the object
    medium_index: float
        Refractive index of the surrounding medium
    effective_fiber_distance: float
        Effective distance between the two trapping fibers relative
        to the medium refractive index [m]. For an open setup, this is
        the physical distance between the fibers. For a closed setup
        (capillary), this distance takes into account the refractive
        indices and thicknesses of the glass capillary and index
        matching gel. For the closed setup, the convenience function
        :func:`ggf.fiber_distance_capillary` can be used.
    mode_field_diameter: float
        The mode field diameter MFD of the fiber used [m]. Note that
        the MFD is dependent on the wavelength used. If the
        manufacturer did not provide a value for the MFD, the MFD
        can be approximated as ``3*wavelenth`` for a single-mode
        fiber.
    power_per_fiber: float
        The laser power coupled into each of the fibers [W]
    wavelength: float
        The laser wavelength used for the trap [m]
    poisson_ratio: float
        The Poisson's ratio of the stretched material. Set this
        to 0.5 for volume conservation.
    n_poly: int
        Number of Legendre polynomials to use for computing the GGF.
        Note that only even Legendre polynomials are used and thus,
        this number is effectively halved.
    lut_path: str or pathlib.Path
        Path to a LUT hdf5 file. If `None`, the internal LUTs are
        used.
    verbose: int
        Increases verbosity

    Returns
    -------
    ggf: float
        Linearly interpolated global geometric factor

    Notes
    -----
    - To avoid invalid values in the look-up table (LUT), such as
      `semi_major < semi_minor` or `object_index < medium_index`,
      the LUT is not built using the exact same keyword arguments
      as this method:

      - `object_index` is stored as
        ``relative_object_index = object_index / medium_index``
      - `semi_major` is stored as
        ``stretch_ratio = (semi_major - semi_minor) / semi_minor``
    - The following keywords are not interpolated in the LUT:

      - `model`
      - `wavelength`: the OS uses a fixed wavelength
      - `mode_field_diameter`: the fiber geometry is fixed
      - `power_per_fiber`: usually fixed for reproducibility
      - `n_poly`: set to a high number (e.g. 120)
    - The following are approximate guiding values for when a keyword
      can be considered linear:

      - stretch_ratio: linear only within interval of 0.004
      - semi_minor: linear only within interval of 0.08µm
      - relative_object_index: linear only within interval of 0.003
      - medium_index: linear only within interval of 0.005
      - poisson_ratio: good linearity
      - power_per_fiber: good linearity
      - effective_fiber_distance: linear only within interval of 15µm
    """
    # convert major_axis to stretch ratio
    stretch_ratio = (semi_major - semi_minor) / semi_minor
    # normalize object index with medium_index
    relative_object_index = object_index / medium_index
    # determine the correct LUT
    kwargs = {"model": model,
              "stretch_ratio": stretch_ratio,
              "semi_minor": semi_minor,
              "relative_object_index": relative_object_index,
              "medium_index": medium_index,
              "effective_fiber_distance": effective_fiber_distance,
              "mode_field_diameter": mode_field_diameter,
              "power_per_fiber": power_per_fiber,
              "wavelength": wavelength,
              "poisson_ratio": poisson_ratio,
              "n_poly": n_poly}
    # get the right LUT
    lut_path = _match_lut(kwargs, lut_path)
    if verbose:
        print("Using LUT path: {}".format(lut_path))
    # get LUT data
    with h5py.File(lut_path, mode="r") as h5:
        values = h5["lut"].value
        meta = dict(h5["lut"].attrs)
    # order of interpolation dimensions
    order = meta["dimension_order"].split(",")
    # grid points
    points = []
    for label in order:
        points.append(np.linspace(meta["{} min".format(label)],
                                  meta["{} max".format(label)],
                                  meta["{} num".format(label)]))
    # interpolation coordinates
    xi = [kwargs[kk] for kk in order]
    # perform interpolation
    ggf = interpn(points=points, values=values, xi=xi)
    return np.asscalar(ggf)



"""Decomposition of stress in Legendre polynomials

To compute the GGF, :func:`ggf.globgeomfact.coeff2ggf` uses the
coefficients of the decomposition of the stress into Legendre
polynomials :math:`P_n(\text{cos}(\theta))`. This example visualizes
the small differences between the original stress and the stress
computed from the Legendre coefficients. This plot is automatically
produced by the original Matlab script *StretcherNStress.m*.

Note that the original Matlab yields different results for the
same set of parameters, because the Poisson's ratio (keyword
argument ``poisson_ratio``) is non-zero;
see `issue #1 <https://github.com/GuckLab/ggf/issues/1>`__.
"""
import matplotlib.pylab as plt
import numpy as np
import percache

from ggf.sci_funcs import legendrePlm
from ggf.stress.boyde2009.core import stress


@percache.Cache("stress_decomposition.cache", livesync=True)
def compute(**kwargs):
    "Locally cached version of ggf.core.stress"
    return stress(**kwargs)


# compute default stress
theta, sigmarr, coeff = compute(ret_legendre_decomp=True,
                                n_points=300)

# compute stress from coefficients
numpoints = theta.size
sigmarr_c = np.zeros((numpoints, 1), dtype=float)
for ii in range(numpoints):
    for jj, cc in enumerate(coeff):
        sigmarr_c[ii] += coeff[jj] * \
            np.real_if_close(legendrePlm(0, jj, np.cos(theta[ii])))

# polar plot
plt.figure(figsize=(8, 8))
ax = plt.subplot(111, projection="polar")
plt.plot(theta, sigmarr, '-b', label="computed stress")
plt.plot(theta + np.pi, sigmarr[::-1], '-b')
plt.plot(theta, sigmarr_c, ':r', label="from Legendre coefficients")
plt.plot(theta + np.pi, sigmarr_c[::-1], ':r')
plt.legend()

plt.tight_layout()
plt.show()



"""Creep compliance analysis

This example uses the contour data of an HL60 cell in the OS to
compute its GGF and creep compliance. The `contour data
<_static/creep_compliance_data.h5>`__ were determined
from `this phase-contrast video <_static/creep_compliance.mp4>`__
(prior to video compression). During stretching, the total laser
power was increased from 0.2W to 1.3W (reflexes due to second
harmonic effects appear as white spots).
"""
import ggf
import h5py
import lmfit
import matplotlib.pylab as plt
import numpy as np


def ellipse_fit(radius, theta):
    '''Fit a centered ellipse to data in polar coordinates

    Parameters
    ----------
    radius: 1d ndarray
        radial coordinates
    theta: 1d ndarray
        angular coordinates [rad]

    Returns
    -------
    a, b: floats
        semi-axes of the ellipse; `a` is aligned with theta=0.
    '''
    def residuals(params, radius, theta):
        a = params["a"].value
        b = params["b"].value
        r = a*b / np.sqrt(a**2 * np.sin(theta)**2 + b**2 * np.cos(theta)**2)
        return r - radius

    parms = lmfit.Parameters()
    parms.add(name="a", value=radius.mean())
    parms.add(name="b", value=radius.mean())

    res = lmfit.minimize(residuals, parms, args=(radius, theta))

    return res.params["a"].value, res.params["b"].value


# load the contour data (stored in polar coordinates)
with h5py.File("data/creep_compliance_data.h5", "r") as h5:
    radius = h5["radius"].value * 1e-6  # [µm] to [m]
    theta = h5["theta"].value
    time = h5["time"].value
    meta = dict(h5.attrs)


factors = np.zeros(len(radius), dtype=float)
semimaj = np.zeros(len(radius), dtype=float)
semimin = np.zeros(len(radius), dtype=float)
strains = np.zeros(len(radius), dtype=float)
complnc = np.zeros(len(radius), dtype=float)

for ii in range(len(radius)):
    # determine semi-major and semi-minor axes
    smaj, smin = ellipse_fit(radius[ii], theta[ii])
    semimaj[ii] = smaj
    semimin[ii] = smin
    # compute GGF
    if (time[ii] > meta["time_stretch_begin [s]"]
            and time[ii] < meta["time_stretch_end [s]"]):
        power_per_fiber = meta["power_per_fiber_stretch [W]"]
        f = ggf.get_ggf(
            model="boyde2009",
            semi_major=smaj,
            semi_minor=smin,
            object_index=meta["object_index"],
            medium_index=meta["medium_index"],
            effective_fiber_distance=meta["effective_fiber_distance [m]"],
            mode_field_diameter=meta["mode_field_diameter [m]"],
            power_per_fiber=power_per_fiber,
            wavelength=meta["wavelength [m]"],
            poisson_ratio=.5)
    else:
        power_per_fiber = meta["power_per_fiber_trap [W]"]
        f = np.nan
    factors[ii] = f

# compute compliance
strains = (semimaj-semimaj[0]) / semimaj[0]
complnc = strains / factors
compl_ival = (time > meta["time_stretch_begin [s]"]) * \
    (time < meta["time_stretch_end [s]"])
stretch_index = np.where(compl_ival)[0][0]
complnc_1 = strains/factors[stretch_index]

# plots
plt.figure(figsize=(8, 7))

ax1 = plt.subplot(221, title="ellipse fit semi-axes")
ax1.plot(time, semimaj*1e6, label="semi-major axis")
ax1.plot(time, semimin*1e6, label="semi-minor axis")
ax1.legend()
ax1.set_xlabel("time [s]")
ax1.set_ylabel("axis radius [µm]")

ax2 = plt.subplot(222, title="GGF")
ax2.plot(time, factors)
ax2.set_xlabel("time [s]")
ax2.set_ylabel("global geometric factor [Pa]")

ax3 = plt.subplot(223, title="strain")
ax3.plot(time, (strains)*100)
ax3.set_xlabel("time [s]")
ax3.set_ylabel("deformation $w(t)/r_0$ [%]")

ax4 = plt.subplot(224, title="creep compliance")
ax4.plot(time[compl_ival], complnc[compl_ival])
ax4.set_xlabel("time [s]")
ax4.set_ylabel("compliance $J(t)$ [Pa⁻¹]")

for ax in [ax1, ax2, ax3, ax4]:
    ax.set_xlim(0, np.round(time.max()))
    ax.axvline(x=meta["time_stretch_begin [s]"], c="r")
    ax.axvline(x=meta["time_stretch_end [s]"], c="r")

plt.tight_layout()
plt.show()



import numpy as np

from .matlab_funcs import lscov, legendre
from .sci_funcs import legendrePlm


def legendre2ggf(coeff, poisson_ratio):
    """Compute the global geometric factor from Legendre coefficients

    The definition of the Legendre coefficients is given in
    :ref:`the theory section <sec_theory_ggf>`.

    Parameters
    ----------
    coeff: 1d ndarray
        Legendre coefficients as defined in :cite:`Lure1964`
    poisson_ratio: float
        Poisson's ratio of the stretched material. Set this
        to 0.5 for volume conservation.

    Returns
    -------
    ggf: float
        Global geometric factor

    Notes
    -----
    All odd Legendre coefficients are assumed to be zero, because the
    stress profile is symmetric with respect to the stretcher axis.
    """
    m = 1 / poisson_ratio

    def Delta(n): return n * (n - 1) + (2 * n + 1) * (m + 1) / m

    def L_n(n): return -1 / Delta(n) * (2 * n + 1) * (n + 1) * (n - 2 + 4 / m)

    def M_n(n): return 1 / Delta(n) * (2 * n + 1) * \
        (n**2 + 2 * n - 1 + 2 / m) * n / (n - 1)
    # Q_n = lambda n: -1/Delta(n) * (2*n+1) * (n + 5 - 4/m)
    # S_n = lambda n: M_n(n) / n
    x = 1  # evaluate displacements at the boundary of the sphere
    theta = 0  # evaluate displacements only on the trapping axis

    # We use the notation: u_r(theta=0, R=radius) / radius = GGF / G
    # Thus, in Lur'e eq. (6.6.8), we move radius and G to the left.
    # ggf = u_r * G / radius

    ggf = 0

    for n, sn in enumerate(coeff):
        if n == 0:
            # n=0 contribution:
            ggf += (m - 2) * sn / (2 * (m + 1))
        elif n % 2:
            if not np.allclose(sn, 0):
                msg = "Odd coeffecient n={} is non-zero: {}".format(n, sn)
                raise ValueError(msg)
        else:
            ggf += 1 / 8 * 2 * sn / (2 * n + 1) \
                * (L_n(n) * x**n + M_n(n) * x**(n - 2)) \
                * np.real_if_close(legendre(n, np.cos(theta))[0][0])

    # Note that u_theta is not considered here!
    return ggf


def stress2legendre(stress, theta, n_poly):
    """Decompose stress into even Legendre Polynomials

    The definition of the Legendre decomposition is given in
    :ref:`the theory section <sec_theory_ggf>`.

    Parameters
    ----------
    stress: 1d ndarray
        Radial stress profile (in imaging plane)
    theta: 1d ndarray
        Polar angles corresponding to `stress`
    n_poly: int
        Number of Legendre polynomials to use

    Returns
    -------
    coeff: 1d ndarray
        Legendre coefficients as defined in :cite:`Lure1964`

    Notes
    -----
    All odd Legendre coefficients are assumed to be zero, because the
    stress profile is symmetric with respect to the stretcher axis.
    Therefore, only `n_poly/2` polynomials are considered.
    """
    # Sigma = Sum_n [Coeff(n) P_n(np.cos(theta))]
    # number of Legendre polynomials used in fit
    nmax = n_poly
    # transfer data from stress plot into pair of corresponding variables
    # [Theta,Sigma]
    numpoints = theta.shape[0]
    theta = theta.reshape(-1, 1)
    sigma = stress.reshape(-1, 1)

    # Write set of linear equations for stresses in terms of Legendre functions
    legmat = np.zeros((numpoints, nmax), dtype=float)
    for ii in range(numpoints):
        # skip odd Legendre Polynomials since stress is an even function
        # (symmetrical)
        for jj in np.arange(nmax)[::2]:
            legmat[ii, jj] = np.real_if_close(
                legendrePlm(0, jj, np.cos(theta[ii])))

    coeff = lscov(legmat, sigma)

    return coeff


def stress2ggf(stress, theta, poisson_ratio, n_poly=120):
    """Compute the GGf from radial stress using Legendre decomposition

    Parameters
    ----------
    stress: 1d ndarray
        Radial stress profile (in imaging plane)
    theta: 1d ndarray
        Polar angles corresponding to `stress`
    poisson_ratio: float
        Poisson's ratio of the stretched material. Set this
        to 0.5 for volume conservation.
    n_poly: int
        Number of Legendre polynomials to use

    Returns
    -------
    ggf: float
        Global geometric factor

    Notes
    -----
    All odd Legendre coefficients are assumed to be zero, because the
    stress profile is symmetric with respect to the stretcher axis.
    Therefore, only `n_poly/2` polynomials are considered.
    """
    coeff = stress2legendre(stress=stress, theta=theta, n_poly=n_poly)

    ggf = legendre2ggf(coeff=coeff, poisson_ratio=poisson_ratio)

    return ggf



"""Special functions translated from Matlab to Python"""
import numpy as np
from scipy import special
from scipy import integrate


def besselh(n, z):
    """Hankel function with k = 1

    Parameters
    ----------
    n: int
        real order
    z: float
        complex argument

    Notes
    -----
    https://de.mathworks.com/help/matlab/ref/besselh.html
    """
    return special.hankel1(n, z)


def besselj(n, z):
    """Bessel function of first kind

    Parameters
    ----------
    n: int
        real order
    z: float
        complex argument

    Notes
    -----
    https://de.mathworks.com/help/matlab/ref/besselj.html
    """
    return special.jv(n, z)


def gammaln(x):
    """Logarithm of the absolute value of the Gamma function

    Notes
    -----
    https://de.mathworks.com/help/matlab/ref/gammaln.html

    See Also
    --------
    scipy.special.gammaln
    """
    return special.gammaln(x)


def legendre(n, x):
    """Associated Legendre functions

    Parameters
    ----------
    n: int
        degree
    x: ndarray of floats
        argument

    Notes
    -----
    x is treated always as a row vector

    The statement legendre(2,0:0.1:0.2) returns the matrix

    =======  =======  =======  =======
    /        x = 0    x = 0.1  x = 0.2
    =======  =======  =======  =======
    m = 0    -0.5000  -0.4850  -0.4400
    m = 1    0        -0.2985  -0.5879
    m = 2    3.0000   2.9700   2.8800
    =======  =======  =======  =======

    Notes
    -----
    https://de.mathworks.com/help/matlab/ref/legendre.html
    """
    x = np.real_if_close(x).flatten()  # flattened
    x = np.array(x, dtype=float)  # warning if not real
    result = np.zeros((x.shape[0], n + 1), dtype=complex)
    for ii in range(x.shape[0]):
        # Gives us row vector
        a = special.lpmn(n, n, x[ii])[0].transpose()[-1]
        result[ii] = a
    return result.transpose()


def lscov(A, B, w=None):
    """Least-squares solution in presence of known covariance

    :math:`A \\cdot x = B`, that is, :math:`x` minimizes
    :math:`(B - A \\cdot x)^T \\cdot \\text{diag}(w) \\cdot (B - A \\cdot x)`.
    The matrix :math:`w` typically contains either counts or inverse
    variances.

    Parameters
    ----------
    A: matrix or 2d ndarray
        input matrix
    B: vector or 1d ndarray
        input vector

    Notes
    --------
    https://de.mathworks.com/help/matlab/ref/lscov.html
    """
    # https://stackoverflow.com/questions/27128688/how-to-use-least-squares-with-weight-matrix-in-python
    # https://de.mathworks.com/help/matlab/ref/lscov.html
    if w is None:
        Aw = A.copy()
        Bw = B.T.copy()
    else:
        W = np.sqrt(np.diag(np.array(w).flatten()))
        Aw = np.dot(W, A)
        Bw = np.dot(B.T, W)

    # set rcond=1e-10 to prevent diverging odd indices in x
    # (problem specific to ggf/stress computation)
    x, residuals, rank, s = np.linalg.lstsq(Aw, Bw.T, rcond=1e-10)
    return np.array(x).flatten()


def quadl(func, a, b):
    """Numerically evaluate integral with scipy QUADPACK quadrature

    Parameters
    ----------
    func: callable
        function to integrate
    a: float
        interval start
    b: float
        interval end

    Notes
    -----
    https://de.mathworks.com/help/matlab/ref/quadl.html
    """
    def wrapreal(x): return func(x).real

    def wrapimag(x): return func(x).imag

    rp = integrate.quad(wrapreal,
                        np.real_if_close(a),
                        np.real_if_close(b))[0]

    ri = integrate.quad(wrapimag,
                        np.real_if_close(a),
                        np.real_if_close(b))[0]

    return rp + 1j * ri



"""Object boundary: stretching and Poisson's ratio

This example illustrates how the parameters Poisson's ratio
:math:`\nu` and stretch ratio :math:`\epsilon` influence
the object boundary used in :func:`ggf.core.stress` and
defined in :func:`ggf.core.boundary`.

Note that the boundary function was not defined correctly prior to
version 0.3.0 (`issue #1 <https://github.com/GuckLab/ggf/issues/1>`__).
Since version 0.3.0, the semi-minor axis is equivalent to the
keyword argument `a` (1 by default).
"""
import numpy as np
import matplotlib.pylab as plt

from ggf.stress.boyde2009.core import boundary

theta = np.linspace(0, 2*np.pi, 300)
costheta = np.cos(theta)

# change epsilon
eps = [.0, .05, .10, .15, .20]
b1s = []
for ep in eps:
    b1s.append(boundary(costheta=costheta,
                        epsilon=ep,
                        nu=.0))

# change Poisson's ratio
nus = [.0, .25, .5]
b2s = []
for nu in nus:
    b2s.append(boundary(costheta=costheta,
                        epsilon=.1,
                        nu=nu))

# plot
plt.figure(figsize=(8, 4))

ax1 = plt.subplot(121, projection="polar")
for ep, bi in zip(eps, b1s):
    ax1.plot(theta, bi, label="ϵ={:.2f}, ν=0".format(ep))
ax1.legend()

ax2 = plt.subplot(122, projection="polar")
for nu, bi in zip(nus, b2s):
    ax2.plot(theta, bi, label="ϵ=.1, ν={:.1f}".format(nu))
ax2.legend()

plt.tight_layout()
plt.show()



from . import core


def get_stress(semi_major, semi_minor, object_index, medium_index,
               effective_fiber_distance=100e6, mode_field_diameter=4.8e-6,
               power_per_fiber=.6, wavelength=1064e-9, n_points=100,
               verbose=False):
    """Wraps around core.stress without poisson_ratio

    See :func:`ggf.stress.get_stress` for parameter descriptions.
    """
    return core.stress(
        object_index=object_index,
        medium_index=medium_index,
        poisson_ratio=0,
        semi_minor=semi_minor,
        stretch_ratio=(semi_major - semi_minor) / (semi_minor),
        wavelength=wavelength,
        beam_waist=mode_field_diameter / 2 / wavelength,
        power_left=power_per_fiber,
        power_right=power_per_fiber,
        dist=effective_fiber_distance / 2,
        n_points=n_points,
        verbose=verbose)



import pathlib

import numpy as np

from ggf.core import legendre2ggf, stress2ggf, stress2legendre


def test_basic():
    rpath = pathlib.Path(__file__).resolve().parent / "data"
    coeff1 = np.loadtxt(str(rpath / "coeff_droplet1.dat"))
    coeff2 = np.loadtxt(str(rpath / "coeff_droplet2.dat"))

    ggf1 = legendre2ggf(coeff1, poisson_ratio=.45)
    ggf2 = legendre2ggf(coeff2, poisson_ratio=.45)

    assert np.allclose(ggf1, 0.7100497580269144)
    assert np.allclose(ggf2, 0.752157535649718)


def test_stress2legendre():
    """Test legendre decomposition of simple cos²(theta) stress model"""
    sigma_0 = 1.341
    theta = np.linspace(0, np.pi, 100, endpoint=True)
    stress = sigma_0 * (np.cos(theta))**2

    coeff = stress2legendre(stress=stress, theta=theta, n_poly=10)

    # Only valid values for n=0 and n=2
    assert np.allclose(coeff[1], 0)
    assert np.allclose(coeff[3:], 0)

    # analytical solution:
    assert np.allclose(coeff[0], 1/3*sigma_0)
    assert np.allclose(coeff[2], 2/3*sigma_0)


def test_stress2ggf():
    sigma_0 = 1.341
    theta = np.linspace(0, np.pi, 100, endpoint=True)
    stress = sigma_0 * (np.cos(theta))**2
    poisson_ratio = 0.45
    n_poly = 10

    ggf = stress2ggf(stress=stress,
                     theta=theta,
                     poisson_ratio=poisson_ratio,
                     n_poly=n_poly)

    # analytical solution (see Ananthakrishnan 2006, Appendix)
    nu = poisson_ratio
    fg = 1 / (2*(1+nu)) \
        * (1/3 * ((1-2*nu) + (4*nu-7)*(1+nu) / (5*nu+7))
           + (-4*nu + 7)*(1+nu) / (5*nu + 7) * 1)

    assert np.allclose(ggf, fg*sigma_0)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



from . import boyde2009
from .geometry import fiber_distance_capillary  # noqa: F401

VALID_MODELS = ["boyde2009"]


def get_stress(model, semi_major, semi_minor, object_index, medium_index,
               effective_fiber_distance=100e-6, mode_field_diameter=3e-6,
               power_per_fiber=.6, wavelength=1064e-9, n_points=100,
               verbose=False):
    """Compute the optical stress profile in the optical stretcher

    Parameters
    ----------
    model: str
        Model to use, one of: `boyde2009`
    semi_major: float
        Semi-major axis of an ellipse fit to the object perimeter [m]
    semi_minor: float
        Semi-minor axis of an ellipse fit to the object perimeter [m]
    object_index: float
        Refractive index of the object
    medium_index: float
        Refractive index of the surrounding medium
    effective_fiber_distance: float
        Effective distance between the two trapping fibers relative
        to the medium refractive index [m]. For an open setup, this is
        the physical distance between the fibers. For a closed setup
        (capillary), this distance takes into account the refractive
        indices and thicknesses of the glass capillary and index
        matching gel. For the closed setup, the convenience function
        :func:`ggf.fiber_distance_capillary` can be used.
    mode_field_diameter: float
        The mode field diameter MFD of the fiber used [m]. Note that
        the MFD is dependent on the wavelength used. If the
        manufacturer did not provide a value for the MFD, the MFD
        can be approximated as ``3*wavelenth`` for a single-mode
        fiber.
    power_per_fiber: float
        The laser power coupled into each of the fibers [W]
    wavelength: float
        The laser wavelength used for the trap [m]
    n_points: int
        Number of points to compute.
    verbose: int
        Increases verbosity

    Returns
    -------
    theta: 1d ndarray of length `n_points`
        Polar angles [rad]
    sigma: 1d ndarray of length `n_points`
        Radial stress profile along `theta` [Pa]
    """
    if model not in VALID_MODELS:
        msg = "`model` must be one of {}, got '{}'".format(VALID_MODELS, model)
        raise ValueError(msg)

    if model == "boyde2009":
        func = boyde2009.get_stress

    return func(
        semi_major=semi_major,
        semi_minor=semi_minor,
        object_index=object_index,
        medium_index=medium_index,
        effective_fiber_distance=effective_fiber_distance,
        mode_field_diameter=mode_field_diameter,
        power_per_fiber=power_per_fiber,
        wavelength=wavelength,
        n_points=n_points,
        verbose=verbose)



"""Radial stresses of a prolate spheroid

This examples computes radial stress profiles for spheroidal
objects in the optical stretcher, reproducing figures (9) and (10)
of :cite:`Boyde2009`.
"""
import matplotlib.pylab as plt
import numpy as np
import percache

from ggf.stress.boyde2009.core import stress


@percache.Cache("stress_reproduced.cache", livesync=True)
def compute(**kwargs):
    "Locally cached version of ggf.core.stress"
    return stress(**kwargs)


# variables from the publication
alpha = 47
wavelength = 1064e-9
radius = alpha * wavelength / (2 * np.pi)

kwargs = {"stretch_ratio": .1,
          "object_index": 1.375,
          "medium_index": 1.335,
          "wavelength": wavelength,
          "beam_waist": 3,
          "semi_minor": radius,
          "power_left": 1,
          "power_right": 1,
          "poisson_ratio": 0,
          "n_points": 200,
          }

kwargs1 = kwargs.copy()
kwargs1["power_right"] = 0
kwargs1["stretch_ratio"] = 0
kwargs1["dist"] = 90e-6

kwargs2 = kwargs.copy()
kwargs2["power_right"] = 0
kwargs2["stretch_ratio"] = .05
kwargs2["dist"] = 90e-6

kwargs3 = kwargs.copy()
kwargs3["power_right"] = 0
kwargs3["stretch_ratio"] = .1
kwargs3["dist"] = 90e-6

kwargs4 = kwargs.copy()
kwargs4["dist"] = 60e-6

kwargs5 = kwargs.copy()
kwargs5["dist"] = 120e-6

kwargs6 = kwargs.copy()
kwargs6["dist"] = 200e-6


# polar plots
plt.figure(figsize=(8, 5))

th1, sigma1 = compute(**kwargs1)
ax1 = plt.subplot(231, projection='polar')
ax1.plot(th1, sigma1, "k")
ax1.plot(th1 + np.pi, sigma1[::-1], "k")

th2, sigma2 = compute(**kwargs2)
ax2 = plt.subplot(232, projection='polar')
ax2.plot(th2, sigma2, "k")
ax2.plot(th2 + np.pi, sigma2[::-1], "k")

th3, sigma3 = compute(**kwargs3)
ax3 = plt.subplot(233, projection='polar')
ax3.plot(th3, sigma3, "k")
ax3.plot(th3 + np.pi, sigma3[::-1], "k")

for ax in [ax1, ax2, ax3]:
    ax.set_rticks([0, 1.5, 3, 4.5])
    ax.set_rlim(0, 4.5)

th4, sigma4 = compute(**kwargs4)
ax4 = plt.subplot(234, projection='polar')
ax4.plot(th4, sigma4, "k")
ax4.plot(th4 + np.pi, sigma4[::-1], "k")
ax4.set_rticks([0, 4, 8, 12])
ax4.set_rlim(0, 12)

th5, sigma5 = compute(**kwargs5)
ax5 = plt.subplot(235, projection='polar')
ax5.plot(th5, sigma5, "k")
ax5.plot(th5 + np.pi, sigma5[::-1], "k")
ax5.set_rticks([0, 1.5, 3, 4.5])
ax5.set_rlim(0, 4.5)

th6, sigma6 = compute(**kwargs6)
ax6 = plt.subplot(236, projection='polar')
ax6.plot(th6, sigma6, "k")
ax6.plot(th6 + np.pi, sigma6[::-1], "k")
ax6.set_rticks([0, 0.6, 1.2, 1.8])
ax6.set_rlim(0, 1.8)

for ax in [ax1, ax2, ax3, ax4, ax5, ax6]:
    ax.set_thetagrids(np.linspace(0, 360, 12, endpoint=False))

plt.tight_layout()
plt.show()



"""GGF as a function of beam waist radius

Note: This example is not reproducing a previously published figure, which
could be due to false simulation parameters.

This example attempts reproduce figure 7 in :cite:`Bellini2012`.
Two implementations of the optical stretcher are described, the assembled
optical stretcher (AOS) and the monolithic optical stretcher (MOS).
In the publication, these have two different mode field diameters (MFD),
6.2µm (AOS) and 9µm (MOS).
The MFD `is equivalent to twice the gaussian beam waist radius
<https://www.rp-photonics.com/mode_radius.html>`_.
In the original publication, the GGF was plotted over the beam waist
diameter (2w₀) and the AOS and MOS mode field diameters were indicated in the
plots. Here, the data are plotted over the beam waist radius (w₀) and half
of the mode field diameters are indicated in the plot. In comparison to the
original figure, it appears as if a factor of two is missing in
the definition of the plotted beam waist radii.
"""
import matplotlib.pylab as plt
import numpy as np
import percache

from ggf.stress.boyde2009 import stress
from ggf import legendre2ggf
from ggf.stress.boyde2009.stretcher import distance_capillary

medium_index = 1.338

dist1 = distance_capillary(gel_dist=5e-6,
                           glass_dist=40e-6,
                           medium_dist=40e-6,
                           gel_index=1.449,
                           glass_index=1.474,
                           medium_index=medium_index)

dist2 = distance_capillary(gel_dist=0,
                           glass_dist=25e-6,
                           medium_dist=50e-6,
                           gel_index=1.449,
                           glass_index=1.474,
                           medium_index=medium_index)

kwargs = {"object_index": 1.37108,
          "medium_index": medium_index,
          "poisson_ratio": 0,
          "semi_minor": 7.51e-6,
          "stretch_ratio": .041,
          "wavelength": 1064e-9,
          "power_left": 1.5,
          "power_right": 1.5,
          "n_points": 181,
          }

kwargs_aos = kwargs.copy()
kwargs_aos.update({"beam_waist": 3,
                   "dist": dist1,
                   })

kwargs_mos = kwargs.copy()
kwargs_mos.update({"beam_waist": 3,
                   "dist": dist2,
                   })


@percache.Cache("ggf_reproduced.cache", livesync=True)
def compute_ggf(**kw):
    coeff = stress(ret_legendre_decomp=True, **kw)[0]
    return legendre2ggf(coeff, poisson_ratio=kw["poisson_ratio"])


aos_ggf = []
mos_ggf = []

w0s = np.linspace(3e-6, 10e-6, 30)
for w0 in w0s:
    kwargs_aos["beam_waist"] = w0/kwargs["wavelength"]
    aos_ggf.append(compute_ggf(**kwargs_aos))
    kwargs_mos["beam_waist"] = w0/kwargs["wavelength"]
    mos_ggf.append(compute_ggf(**kwargs_mos))

plt.figure(figsize=(8, 4))
ax = plt.subplot(111)
ax.plot(w0s[:len(aos_ggf)]*1e6, aos_ggf,
        marker="^", color="red", label="AOS", lw=2)
ax.plot(w0s[:len(aos_ggf)]*1e6, mos_ggf,
        marker="d", color="blue", label="MOS", lw=2)
ax.axvline(6.2/2, color="red", ls=":", label="AOS beam waist radius")
ax.axvline(9/2, color="blue", ls=":", label="MOS beam waist radius")
ax.legend()
ax.set_ylabel("GGF [Pa]")
ax.set_xlabel("gaussian beam waist radius w₀ [µm]")
ax.grid()

plt.tight_layout()
plt.show()



import numpy as np
import pytest

import ggf


@pytest.mark.xfail
def test_exact_cell():
    """Test the exact value on a LUT grid point"""
    stretch_ratio = 0.065
    semi_minor = 6.724137931034484e-06
    semi_major = semi_minor * (stretch_ratio + 1)
    f = ggf.get_ggf(model="boyde2009",
                    semi_major=semi_major,
                    semi_minor=semi_minor,
                    object_index=1.333*1.025,
                    medium_index=1.333,
                    effective_fiber_distance=175e-6,
                    mode_field_diameter=4.8e-6,
                    power_per_fiber=.65,
                    wavelength=780e-9,
                    poisson_ratio=.5,
                    n_poly=120,
                    use_lut=True)
    exact = 0.7711334992513761
    assert np.allclose(exact, f, rtol=0, atol=1e-10)


@pytest.mark.xfail
def test_exact_fus():
    """Test the exact value on a LUT grid point"""
    f = ggf.get_ggf(model="boyde2009",
                    semi_major=2.8e-6,
                    semi_minor=2.8e-6,
                    object_index=1.340*1.045,
                    medium_index=1.340,
                    effective_fiber_distance=170e-6,
                    mode_field_diameter=4.8e-6,
                    power_per_fiber=.65,
                    wavelength=780e-9,
                    poisson_ratio=.4,
                    n_poly=120,
                    use_lut=True)
    exact = 1.23706977246589
    assert np.allclose(exact, f, rtol=0, atol=1e-10)


@pytest.mark.xfail
def test_basic_fus_semi_major():
    f = ggf.get_ggf(model="boyde2009",
                    semi_major=2.8e-6*1.0018181818181818182,
                    semi_minor=2.8e-6,
                    object_index=1.340*1.045,
                    medium_index=1.340,
                    effective_fiber_distance=170e-6,
                    mode_field_diameter=4.8e-6,
                    power_per_fiber=.65,
                    wavelength=780e-9,
                    poisson_ratio=.4,
                    n_poly=None,
                    use_lut=True)
    exact = 1.237550828156066
    assert np.allclose(exact, f, rtol=0, atol=1e-5)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import numpy as np
import pathlib

from .core import legendre2ggf, stress2ggf, stress2legendre  # noqa: F401
from . import lut
from . import stress
from .stress.geometry import fiber_distance_capillary  # noqa: F401


def get_ggf(model, semi_major, semi_minor, object_index, medium_index,
            effective_fiber_distance=100e-6, mode_field_diameter=3e-6,
            power_per_fiber=.6, wavelength=1064e-9, poisson_ratio=0.5,
            n_poly=120, use_lut=None, verbose=False):
    """Model the global geometric factor

    Parameters
    ----------
    model: str
        Model to use, one of: `boyde2009`
    semi_major: float
        Semi-major axis of an ellipse fit to the object perimeter [m]
    semi_minor: float
        Semi-minor axis of an ellipse fit to the object perimeter [m]
    object_index: float
        Refractive index of the object
    medium_index: float
        Refractive index of the surrounding medium
    effective_fiber_distance: float
        Effective distance between the two trapping fibers relative
        to the medium refractive index [m]. For an open setup, this is
        the physical distance between the fibers. For a closed setup
        (capillary), this distance takes into account the refractive
        indices and thicknesses of the glass capillary and index
        matching gel. For the closed setup, the convenience function
        :func:`ggf.fiber_distance_capillary` can be used.
    mode_field_diameter: float
        The mode field diameter MFD of the fiber used [m]. Note that
        the MFD is dependent on the wavelength used. If the
        manufacturer did not provide a value for the MFD, the MFD
        can be approximated as ``3*wavelenth`` for a single-mode
        fiber.
    power_per_fiber: float
        The laser power coupled into each of the fibers [W]
    wavelength: float
        The laser wavelength used for the trap [m]
    poisson_ratio: float
        The Poisson's ratio of the stretched material. Set this
        to 0.5 for volume conservation.
    n_poly: int
        Number of Legendre polynomials to use for computing the GGF.
        Note that only even Legendre polynomials are used and thus,
        this number is effectively halved. To reproduce the GGF as
        computed with the Boyde2009 Matlab script, set this value
        to `None`.
    use_lut: None, str, pathlib.Path or bool
        Use look-up tables to compute the GGF. If set to `None`,
        the internal LUTs will be used or the GGF is computed if
        it cannot be found in a LUT. If `True`, the internal LUTs
        will be used and a `NotInLUTError` will be raised if the
        GGF cannot be found there. Alternatively, a path to a
        LUT hdf5 file can be passed.
    verbose: int
        Increases verbosity

    Returns
    -------
    ggf: float
        global geometric factor
    """
    if use_lut not in [False]:
        try:
            if isinstance(use_lut, (str, pathlib.Path)):
                lut_path = use_lut
            else:
                lut_path = None
            ggf = lut.get_ggf_lut(
                model=model,
                semi_major=semi_major,
                semi_minor=semi_minor,
                object_index=object_index,
                medium_index=medium_index,
                effective_fiber_distance=effective_fiber_distance,
                mode_field_diameter=mode_field_diameter,
                power_per_fiber=power_per_fiber,
                wavelength=wavelength,
                poisson_ratio=poisson_ratio,
                n_poly=n_poly,
                lut_path=lut_path,
                verbose=verbose)
        except lut.NotInLUTError:
            if use_lut:  # user specifically defined to ONLY use LUT
                raise
            else:  # force computation of the GGF
                ggf = get_ggf(
                    model=model,
                    semi_major=semi_major,
                    semi_minor=semi_minor,
                    object_index=object_index,
                    medium_index=medium_index,
                    effective_fiber_distance=effective_fiber_distance,
                    mode_field_diameter=mode_field_diameter,
                    power_per_fiber=power_per_fiber,
                    wavelength=wavelength,
                    poisson_ratio=poisson_ratio,
                    n_poly=n_poly,
                    use_lut=False,
                    verbose=verbose)
    else:
        theta, sigma = stress.get_stress(
            model=model,
            semi_major=semi_major,
            semi_minor=semi_minor,
            object_index=object_index,
            medium_index=medium_index,
            effective_fiber_distance=effective_fiber_distance,
            mode_field_diameter=mode_field_diameter,
            power_per_fiber=power_per_fiber,
            wavelength=wavelength,
            verbose=verbose)

        if n_poly is None:
            # number of orders (estimate from Boyde 2009)
            alpha = semi_minor * 2 * np.pi / wavelength  # size parameter
            n_poly = np.int(np.round(2 + alpha + 4 * (alpha)**(1 / 3) + 10))
        else:
            n_poly = int(np.round(n_poly))

        ggf = stress2ggf(stress=sigma,
                         theta=theta,
                         poisson_ratio=poisson_ratio,
                         n_poly=n_poly)

    return ggf



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from os.path import dirname, realpath, exists
from setuptools import setup, find_packages
import sys


author = u"Paul Müller"
authors = [author, "Gheorghe Cojoc"]
description = 'global geometric factors and corresponding stresses of the optical stretcher '
name = 'ggf'
year = "2018"

sys.path.insert(0, realpath(dirname(__file__))+"/"+name)
from _version import version

if __name__ == "__main__":
    setup(
        name=name,
        author=author,
        author_email='dev@craban.de',
        url='https://github.com/GuckLab/ggf',
        version=version,
        packages=find_packages(),
        package_dir={name: name},
        include_package_data=True,
        license="MIT",
        description=description,
        long_description=open('README.rst').read() if exists('README.rst') else '',
        install_requires=["h5py>=2.7.0"
                          "numpy>=1.9.0",
                          "scipy>=0.18.0",
                          ],
        setup_requires=['pytest-runner'],
        tests_require=["pytest"],
        python_requires='>=3.6, <4',
        keywords=["optical stretcher",
                  "global geometric factor",
                  ],
        classifiers= [
            'Operating System :: OS Independent',
            'Programming Language :: Python :: 3',
            'Intended Audience :: Science/Research'
                     ],
        platforms=['ALL'],
        )



# flake8: noqa
"""Computation of the global geometric factor"""
import numpy as np

from ...matlab_funcs import lscov, legendre


def coeff2ggf(coeff, poisson_ratio=.45):
    """Compute the global geometric factor from stress coefficients

    The radial displacements of an elastic sphere can be expressed in
    terms of Legendre polynomials (see :cite:`Lure1964` equation 6.2.9)
    whose coefficients are computed from the Legendre decomposition of
    the radial stress.

    Notes
    -----

    - For a :math:`\\sigma_0 \cos^n(\\theta)` stress profile,
      the GGF already includes the peak stress according to:

      .. math::

          \\text{GGF} = \\sigma_0 F_\\text{G}.

    - This is a conversion of the Matlab script GGF.m to Python. The
      code solves a linear system of equations to determine all
      Legendre coefficients. The new implementation in
      :func:`ggf.legendre2ggf` uses the direct solution and
      thus should be preferred.
    """
    # Parameters used in the program

    Theta1 = 0
    # Nu =input('enter the value of nu (Poisson s ratio) \n')
    Nu = poisson_ratio
    ##

    SigmaMat = coeff
    n = SigmaMat.size - 1
    # amendement due to the correction of the paper by Ananthakrishnan
    SigmaMatamended = SigmaMat / 2
    Sigmatot = np.zeros((2 * n + 2, 1))

    for k in range(n):
        Sigmatot[k] = SigmaMatamended[k]
        #Sigmatot[k+n+1] = 0

    Mat = np.zeros((2 * n + 2, 2 * n + 2))
    r = 1  # r is set to 1 for the calculation.
    N = n

    for d in range(n + 1):
        Mat[d, d] = (r**d) * (d + 1) * (d**2 - d - 2 - 2 * Nu)
        Mat[d, d + N + 1] = (r**(d - 2)) * d * (d - 1)

    for d in range(2, n):
        # i pour le coefficient et j pour la ligne de sigma pour le coefficient
        # P(d)
        Mat[d + N, d + 1] = (r**(d + 1)) * (-1 + 2 * (d + 1) + (d + 1)**2 +
                                            2 * Nu) * (-(d + 1)) * (d + 2) / (1 + 2 * (d + 1))
        Mat[d + N, d - 1] = (r**(d - 1)) * (-1 + 2 * (d - 1) + ((d - 1)**2) +
                                            2 * Nu) * (d) * (d - 1) / (1 + 2 * (d - 1))
        Mat[d + N, d + N + 1 + 1] = (r**(d - 1)) * \
            (d) * (-(d + 1)) * (d + 2) / (1 + 2 * (d + 1))
        Mat[d + N, d + N] = (r**(d - 3)) * (d - 2) * \
            (d) * (d - 1) / (1 + 2 * (d - 1))

    d = 0
    Mat[d + N, d + 1] = (r**(d + 1)) * (-1 + 2 * (d + 1) + (d + 1)**2 + 2 * Nu) * \
        (-(d + 1)) * (d + 2) / (1 + 2 * (d + 1))
    Mat[d + N, d + N + 1 + 1] = (r**(d - 1)) * \
        (d) * (-(d + 1)) * (d + 2) / (1 + 2 * (d + 1))

    d = 1
    Mat[d + N, d + 1] = (r**(d + 1)) * (-1 + 2 * (d + 1) + ((d + 1)**2) + 2 * Nu) * \
        (-(d + 1)) * (d + 2) / (1 + 2 * (d + 1))
    Mat[d + N, d + N + 1 + 1] = (r**(d - 1)) * \
        (d) * (-(d + 1)) * (d + 2) / (1 + 2 * (d + 1))

    d = n
    Mat[d + N, d - 1] = (r**(d - 1)) * (-1 + 2 * (d - 1) + ((d - 1)**2) + 2 * Nu) * \
        (d) * (d - 1) / (1 + 2 * (d - 1))
    Mat[d + N, d + N] = (r**(d - 3)) * (d - 2) * (d) * \
        (d - 1) / (1 + 2 * (d - 1))

    d = n + 1
    Mat[d + N, d - 1] = (r**(d - 1)) * (-1 + 2 * (d - 1) + ((d - 1)**2) + 2 * Nu) * \
        (d) * (d - 1) / (1 + 2 * (d - 1))
    Mat[d + N, d + N] = (r**(d - 3)) * (d - 2) * (d) * \
        (d - 1) / (1 + 2 * (d - 1))

    Mat1 = Mat
    sol = lscov(Mat1, Sigmatot)

    PL = np.zeros(n + 1, dtype=float)
    w0 = np.zeros(n + 1, dtype=float)

    # Calculation of the Global Geometrical Factor including sigma0
    ww = 0
    for k in range(0, n):
        PLA = legendre(k, np.cos(Theta1))
        PL[k] = np.real_if_close(PLA[0][0])  # legendre polynomial with m=0!
        w0[k] = ((sol[k]) * (r**(k + 1)) * (k + 1) * (k - 2 + 4 * Nu) +
                 (sol[n + k + 1]) * r**(k - 1) * k) * PL[k]
        ww = ww + w0[k]

    GF = ww

    return GF





def fiber_distance_capillary(gel_thickness=2e-6, glass_thickness=40e-6,
                             channel_width=40e-6, gel_index=1.449,
                             glass_index=1.474, medium_index=1.335):
    """Effective distance between the two optical fibers

    When the optical stretcher is combined with a microfluidic
    channel (closed setup), then the effective distance between
    the two optical fibers (with the the stretched object at the
    channel center) is defined by the refractive indices of
    the optical components: index matching gel between fiber and
    channel wall, microfluidic glass channel wall, and medium
    inside the channel.

    Parameters
    ----------
    gel_thickness: float
        Thickness of index matching gel (distance between fiber
        and glass wall) [m]
    glass_thickness: float
        Thickness of glass wall [m]
    channel_width: float
        Width of the microfluidic channed [m]
    gel_index: float
        Refractive index of index matching gel
    glass_index: float
        Refractive index of channel glass wall
    medium_index: float
        Refractive index of index medium inside channel

    Returns
    -------
    eff_dist: float
        Effective distance between the fibers

    Notes
    -----
    The effective distance is computed relative to the medium,
    i.e. if `gel_index` == `glass_index` == `medium_index`, then
    `eff_dist` = 2*`gel_dist` + 2*`glass_dist` + `channel_width`.
    """
    eff_dist = 2 * medium_index / gel_index * gel_thickness \
        + 2 * medium_index / glass_index * glass_thickness \
        + channel_width
    return eff_dist
